from typing import List
from typing import Optional

from evidently.base_metric import InputData
from evidently.test_preset.test_preset import TestPreset
from evidently.tests import TestAllColumnsMostCommonValueShare
from evidently.tests import TestAllColumnsShareOfMissingValues
from evidently.tests import TestHighlyCorrelatedColumns
from evidently.tests import TestNumberOfConstantColumns
from evidently.tests import TestNumberOfDuplicatedColumns
from evidently.tests import TestNumberOfDuplicatedRows
from evidently.utils.data_operations import DatasetColumns


class DataQualityTestPreset(TestPreset):
    """
    Data Quality tests.

    Contains tests:
    - `TestAllColumnsShareOfMissingValues`
    - `TestAllColumnsMostCommonValueShare`
    - `TestNumberOfConstantColumns`
    - `TestNumberOfDuplicatedColumns`
    - `TestNumberOfDuplicatedRows`
    - `TestHighlyCorrelatedColumns`
    """

    columns: Optional[List[str]]

    def __init__(
        self,
        columns: Optional[List[str]] = None,
    ):
        super().__init__()
        self.columns = columns

    def generate_tests(self, data: InputData, columns: DatasetColumns):
        return [
            TestAllColumnsShareOfMissingValues(columns=self.columns),
            TestAllColumnsMostCommonValueShare(columns=self.columns),
            TestNumberOfConstantColumns(),
            TestNumberOfDuplicatedColumns(),
            TestNumberOfDuplicatedRows(),
            TestHighlyCorrelatedColumns(),
        ]



import dataclasses
import json
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
from pandas.api.types import is_string_dtype
from plotly.subplots import make_subplots

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.classification_performance import PredictionData
from evidently.calculations.classification_performance import get_prediction_data
from evidently.features.non_letter_character_percentage_feature import NonLetterCharacterPercentage
from evidently.features.OOV_words_percentage_feature import OOVWordsPercentage
from evidently.features.text_length_feature import TextLength
from evidently.model.widget import AdditionalGraphInfo
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.utils.data_operations import process_columns
from evidently.utils.data_preprocessing import DataDefinition


@dataclasses.dataclass
class TargetByFeaturesTableResults:
    current_plot_data: pd.DataFrame
    reference_plot_data: pd.DataFrame
    target_name: Optional[str]
    curr_predictions: Optional[PredictionData]
    ref_predictions: Optional[PredictionData]
    columns: List[str]
    task: str


class TargetByFeaturesTable(Metric[TargetByFeaturesTableResults]):
    columns: Optional[List[str]]
    text_features_gen: Optional[
        Dict[str, Dict[str, Union[TextLength, NonLetterCharacterPercentage, OOVWordsPercentage]]]
    ]

    def __init__(self, columns: Optional[List[str]] = None):
        self.columns = columns
        self.text_features_gen = None

    def required_features(self, data_definition: DataDefinition):
        if len(data_definition.get_columns("text_features")) > 0:
            text_cols = [col.column_name for col in data_definition.get_columns("text_features")]
            text_features_gen = {}
            text_features_gen_result = []
            for col in text_cols:
                col_dict: Dict[str, Union[TextLength, NonLetterCharacterPercentage, OOVWordsPercentage]] = {}
                col_dict[f"{col}: Text Length"] = TextLength(col)
                col_dict[f"{col}: Non Letter Character %"] = NonLetterCharacterPercentage(col)
                col_dict[f"{col}: OOV %"] = OOVWordsPercentage(col)

                text_features_gen_result += [
                    col_dict[f"{col}: Text Length"],
                    col_dict[f"{col}: Non Letter Character %"],
                    col_dict[f"{col}: OOV %"],
                ]
                text_features_gen[col] = col_dict
            self.text_features_gen = text_features_gen

            return text_features_gen_result
        else:
            return []

    def get_parameters(self) -> tuple:
        return ()

    def calculate(self, data: InputData) -> TargetByFeaturesTableResults:
        dataset_columns = process_columns(data.current_data, data.column_mapping)
        target_name = dataset_columns.utility_columns.target
        prediction_name = dataset_columns.utility_columns.prediction
        if target_name is None and prediction_name is None:
            raise ValueError("The columns 'target' or 'prediction' should be present")
        if data.reference_data is None:
            raise ValueError("Reference data should be present")
        curr_df = data.current_data.copy()
        ref_df = data.reference_data.copy()
        curr_predictions = None
        ref_predictions = None
        if prediction_name is not None:
            curr_predictions = get_prediction_data(data.current_data, dataset_columns, data.column_mapping.pos_label)
            ref_predictions = get_prediction_data(data.reference_data, dataset_columns, data.column_mapping.pos_label)

        if self.columns is None:
            columns = (
                dataset_columns.num_feature_names
                + dataset_columns.cat_feature_names
                + dataset_columns.text_feature_names
            )
        else:
            columns = list(
                np.intersect1d(
                    self.columns,
                    (
                        dataset_columns.num_feature_names
                        + dataset_columns.cat_feature_names
                        + dataset_columns.text_feature_names
                    ),
                )
            )
        if data.column_mapping.task is not None:
            task = data.column_mapping.task
        else:
            if target_name is not None:
                if curr_df[target_name].nunique() < 5 or is_string_dtype(curr_df[target_name]):
                    task = "classification"
                else:
                    task = "regression"
            else:
                raise ValueError("Task parameter of column_mapping should be specified")
        # process text columns
        if (
            self.text_features_gen is not None
            and len(np.intersect1d(list(self.text_features_gen.keys()), columns)) >= 1
        ):
            for col in np.intersect1d(list(self.text_features_gen.keys()), columns):
                columns += list(self.text_features_gen[col].keys())
                columns.remove(col)
                curr_text_df = pd.concat(
                    [data.get_current_column(x.feature_name()) for x in list(self.text_features_gen[col].values())],
                    axis=1,
                )
                curr_text_df.columns = list(self.text_features_gen[col].keys())
                curr_df = pd.concat([curr_df.reset_index(drop=True), curr_text_df.reset_index(drop=True)], axis=1)

                if ref_df is not None:
                    ref_text_df = pd.concat(
                        [
                            data.get_reference_column(x.feature_name())
                            for x in list(self.text_features_gen[col].values())
                        ],
                        axis=1,
                    )
                    ref_text_df.columns = list(self.text_features_gen[col].keys())
                    ref_df = pd.concat([ref_df.reset_index(drop=True), ref_text_df.reset_index(drop=True)], axis=1)

        return TargetByFeaturesTableResults(
            current_plot_data=curr_df,
            reference_plot_data=ref_df,
            curr_predictions=curr_predictions,
            ref_predictions=ref_predictions,
            columns=columns,
            target_name=target_name,
            task=task,
        )


@default_renderer(wrap_type=TargetByFeaturesTable)
class TargetByFeaturesTableRenderer(MetricRenderer):
    def render_json(self, obj: TargetByFeaturesTable) -> dict:
        return {}

    def render_html(self, obj: TargetByFeaturesTable) -> List[BaseWidgetInfo]:
        result = obj.get_result()
        current_data = result.current_plot_data
        reference_data = result.reference_plot_data
        target_name = result.target_name
        curr_predictions = result.curr_predictions
        ref_predictions = result.ref_predictions
        columns = result.columns
        task = result.task

        if curr_predictions is not None and ref_predictions is not None:
            current_data["prediction_labels"] = curr_predictions.predictions.values
            reference_data["prediction_labels"] = ref_predictions.predictions.values

        additional_graphs_data = []
        params_data = []

        for feature_name in columns:
            # add data for table in params
            parts = []

            if target_name is not None:
                parts.append({"title": "Target", "id": feature_name + "_target_values"})
                if task == "regression":
                    target_fig = self._get_regression_fig(feature_name, target_name, current_data, reference_data)
                else:
                    target_fig = self._get_classification_fig(feature_name, target_name, current_data, reference_data)

                target_fig_json = json.loads(target_fig.to_json())

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_target_values",
                        {"data": target_fig_json["data"], "layout": target_fig_json["layout"]},
                    )
                )

            if curr_predictions is not None:
                parts.append({"title": "Prediction", "id": feature_name + "_prediction_values"})
                if task == "regression":
                    preds_fig = self._get_regression_fig(
                        feature_name, "prediction_labels", current_data, reference_data
                    )
                else:
                    preds_fig = self._get_classification_fig(
                        feature_name, "prediction_labels", current_data, reference_data
                    )
                preds_fig_json = json.loads(preds_fig.to_json())

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_prediction_values",
                        {"data": preds_fig_json["data"], "layout": preds_fig_json["layout"]},
                    )
                )

            params_data.append(
                {
                    "details": {
                        "parts": parts,
                        "insights": [],
                    },
                    "f1": feature_name,
                }
            )
        return [
            BaseWidgetInfo(
                title="Target (Prediction) Behavior By Feature",
                type="big_table",
                size=2,
                params={
                    "rowsPerPage": min(len(columns), 10),
                    "columns": [{"title": "Feature", "field": "f1"}],
                    "data": params_data,
                },
                additionalGraphs=additional_graphs_data,
            )
        ]

    def _get_regression_fig(self, feature_name: str, main_column: str, curr_data: pd.DataFrame, ref_data: pd.DataFrame):
        fig = make_subplots(rows=1, cols=2, subplot_titles=("Current", "Reference"), shared_yaxes=True)
        fig.add_trace(
            go.Scattergl(
                x=curr_data[feature_name],
                y=curr_data[main_column],
                mode="markers",
                name="current",
                marker=dict(size=6, color=self.color_options.primary_color),
            ),
            row=1,
            col=1,
        )

        fig.add_trace(
            go.Scatter(
                x=ref_data[feature_name],
                y=ref_data[main_column],
                mode="markers",
                name="reference",
                marker=dict(size=6, color=self.color_options.secondary_color),
            ),
            row=1,
            col=2,
        )
        fig.update_xaxes(title_text=feature_name, showgrid=True, row=1, col=1)
        fig.update_xaxes(title_text=feature_name, showgrid=True, row=1, col=2)
        fig.update_yaxes(title_text=main_column, showgrid=True, row=1, col=1)

        return fig

    def _get_classification_fig(
        self, feature_name: str, main_column: str, curr_data: pd.DataFrame, ref_data: pd.DataFrame
    ):
        curr = curr_data.copy()
        ref = ref_data.copy()
        ref["dataset"] = "Reference"
        curr["dataset"] = "Current"
        merged_data = pd.concat([ref, curr])
        fig = px.histogram(
            merged_data,
            x=feature_name,
            color=main_column,
            facet_col="dataset",
            barmode="overlay",
            category_orders={"dataset": ["Current", "Reference"]},
        )

        return fig



import dataclasses
from typing import List
from typing import Optional

import pandas as pd
from sklearn import metrics

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.classification_performance import PredictionData
from evidently.calculations.classification_performance import get_prediction_data
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import TabData
from evidently.renderers.html_widgets import get_roc_auc_tab_data
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import widget_tabs
from evidently.utils.data_operations import process_columns


@dataclasses.dataclass
class ClassificationRocCurveResults:
    current_roc_curve: Optional[dict] = None
    reference_roc_curve: Optional[dict] = None


class ClassificationRocCurve(Metric[ClassificationRocCurveResults]):
    def calculate(self, data: InputData) -> ClassificationRocCurveResults:
        dataset_columns = process_columns(data.current_data, data.column_mapping)
        target_name = dataset_columns.utility_columns.target
        prediction_name = dataset_columns.utility_columns.prediction
        if target_name is None or prediction_name is None:
            raise ValueError("The columns 'target' and 'prediction' columns should be present")
        curr_predictions = get_prediction_data(data.current_data, dataset_columns, data.column_mapping.pos_label)
        if curr_predictions.prediction_probas is None:
            raise ValueError("Roc Curve can be calculated only on binary probabilistic predictions")
        curr_roc_curve = self.calculate_metrics(data.current_data[target_name], curr_predictions)
        ref_roc_curve = None
        if data.reference_data is not None:
            ref_predictions = get_prediction_data(data.reference_data, dataset_columns, data.column_mapping.pos_label)
            ref_roc_curve = self.calculate_metrics(data.reference_data[target_name], ref_predictions)
        return ClassificationRocCurveResults(
            current_roc_curve=curr_roc_curve,
            reference_roc_curve=ref_roc_curve,
        )

    def calculate_metrics(self, target_data: pd.Series, prediction: PredictionData):
        labels = prediction.labels
        if prediction.prediction_probas is None:
            raise ValueError("Roc Curve can be calculated only on binary probabilistic predictions")
        binaraized_target = (target_data.values.reshape(-1, 1) == labels).astype(int)
        roc_curve = {}
        if len(labels) <= 2:
            binaraized_target = pd.DataFrame(binaraized_target[:, 0])
            binaraized_target.columns = ["target"]

            fpr, tpr, thrs = metrics.roc_curve(binaraized_target, prediction.prediction_probas.iloc[:, 0])
            roc_curve[prediction.prediction_probas.columns[0]] = {
                "fpr": fpr.tolist(),
                "tpr": tpr.tolist(),
                "thrs": thrs.tolist(),
            }
        else:
            binaraized_target = pd.DataFrame(binaraized_target)
            binaraized_target.columns = labels

            for label in labels:
                fpr, tpr, thrs = metrics.roc_curve(binaraized_target[label], prediction.prediction_probas[label])
                roc_curve[label] = {
                    "fpr": fpr.tolist(),
                    "tpr": tpr.tolist(),
                    "thrs": thrs.tolist(),
                }
        return roc_curve


@default_renderer(wrap_type=ClassificationRocCurve)
class ClassificationRocCurveRenderer(MetricRenderer):
    def render_json(self, obj: ClassificationRocCurve) -> dict:
        return {}

    def render_html(self, obj: ClassificationRocCurve) -> List[BaseWidgetInfo]:
        current_roc_curve = obj.get_result().current_roc_curve
        reference_roc_curve = obj.get_result().reference_roc_curve
        if current_roc_curve is None:
            return []

        tab_data = get_roc_auc_tab_data(current_roc_curve, reference_roc_curve, color_options=self.color_options)
        if len(tab_data) == 1:
            return [header_text(label="ROC Curve"), tab_data[0][1]]
        tabs = [TabData(name, widget) for name, widget in tab_data]
        return [header_text(label="ROC Curve"), widget_tabs(title="", tabs=tabs)]



#!/usr/bin/env python
# coding: utf-8

from .cat_target_drift_profile_section import CatTargetDriftProfileSection
from .classification_performance_profile_section import (
    ClassificationPerformanceProfileSection,
)
from .data_drift_profile_section import DataDriftProfileSection
from .data_quality_profile_section import DataQualityProfileSection
from .num_target_drift_profile_section import NumTargetDriftProfileSection
from .prob_classification_performance_profile_section import (
    ProbClassificationPerformanceProfileSection,
)
from .regression_performance_profile_section import RegressionPerformanceProfileSection



#!/usr/bin/env python
# coding: utf-8

from typing import Optional

import pandas as pd
import plotly.graph_objs as go

from evidently import ColumnMapping
from evidently.analyzers.num_target_drift_analyzer import NumTargetDriftAnalyzer
from evidently.dashboard.widgets.utils import fig_to_json
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions


class NumOutputCorrWidget(Widget):
    def __init__(self, title: str, kind: str = "target"):
        super().__init__(title)
        self.title = title
        self.kind = kind  # target or prediction

    def analyzers(self):
        return [NumTargetDriftAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        color_options = self.options_provider.get(ColorOptions)
        results = NumTargetDriftAnalyzer.get_results(analyzers_results)

        if self.kind == "target":
            if results.columns.utility_columns.target is None:
                return None

            metrics = results.target_metrics

        elif self.kind == "prediction":
            if results.columns.utility_columns.prediction is None:
                return None

            metrics = results.prediction_metrics

        else:
            raise ValueError(
                f"Widget [{self.title}] requires 'target' or 'prediction' kind parameter value"
            )

        if metrics is None:
            return None

        # calculate corr
        ref_output_corr = metrics.reference_correlations

        if ref_output_corr is None:
            return None

        current_output_corr = metrics.current_correlations

        if current_output_corr is None:
            return None

        # plot output correlations
        output_corr = go.Figure()

        output_corr.add_trace(
            go.Bar(
                y=list(ref_output_corr.values()),
                x=list(ref_output_corr.keys()),
                marker_color=color_options.get_reference_data_color(),
                name="Reference",
            )
        )

        output_corr.add_trace(
            go.Bar(
                y=list(current_output_corr.values()),
                x=list(current_output_corr.keys()),
                marker_color=color_options.get_current_data_color(),
                name="Current",
            )
        )

        output_corr.update_layout(
            xaxis_title="Features",
            yaxis_title="Correlation",
            yaxis=dict(range=(-1, 1), showticklabels=True),
        )

        # output_corr_json = json.loads(output_corr.to_json())
        output_corr_json = fig_to_json(output_corr)

        return BaseWidgetInfo(
            title=self.title,
            type="big_graph",
            size=1,
            params={
                "data": output_corr_json["data"],
                "layout": output_corr_json["layout"],
            },
        )



import dataclasses
from typing import List
from typing import Optional
from typing import Union

from evidently.base_metric import InputData
from evidently.calculations.classification_performance import ConfusionMatrix
from evidently.calculations.classification_performance import calculate_matrix
from evidently.metrics.classification_performance.base_classification_metric import ThresholdClassificationMetric
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import plotly_figure
from evidently.utils.visualizations import plot_conf_mtrx

DEFAULT_THRESHOLD = 0.5


@dataclasses.dataclass
class ClassificationConfusionMatrixResult:
    current_matrix: ConfusionMatrix
    reference_matrix: Optional[ConfusionMatrix]


class ClassificationConfusionMatrix(ThresholdClassificationMetric[ClassificationConfusionMatrixResult]):
    probas_threshold: Optional[float]
    k: Optional[Union[float, int]]

    def __init__(
        self,
        probas_threshold: Optional[float] = None,
        k: Optional[Union[float, int]] = None,
    ):
        super().__init__(probas_threshold=probas_threshold, k=k)

    def calculate(self, data: InputData) -> ClassificationConfusionMatrixResult:
        current_target_data, current_pred = self.get_target_prediction_data(data.current_data, data.column_mapping)

        current_results = calculate_matrix(
            current_target_data,
            current_pred.predictions,
            current_pred.labels,
        )

        reference_results = None
        if data.reference_data is not None:
            ref_target_data, ref_pred = self.get_target_prediction_data(data.reference_data, data.column_mapping)
            reference_results = calculate_matrix(
                ref_target_data,
                ref_pred.predictions,
                ref_pred.labels,
            )

        return ClassificationConfusionMatrixResult(
            current_matrix=current_results,
            reference_matrix=reference_results,
        )


@default_renderer(wrap_type=ClassificationConfusionMatrix)
class ClassificationConfusionMatrixRenderer(MetricRenderer):
    def render_json(self, obj: ClassificationConfusionMatrix) -> dict:
        return {}

    def render_html(self, obj: ClassificationConfusionMatrix) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        fig = plot_conf_mtrx(metric_result.current_matrix, metric_result.reference_matrix)
        fig.for_each_xaxis(lambda axis: axis.update(title_text="Predicted Value"))
        fig.update_layout(yaxis_title="Actual Value")
        return [header_text(label="Confusion Matrix"), plotly_figure(figure=fig, title="")]



"""Chisquare test of two samples.

Name: "chisquare"

Import:

    >>> from evidently.calculations.stattests import chi_stat_test

Properties:
- only for categorical features
- returns p-value

Example:
    Using by object:

    >>> from evidently.options import DataDriftOptions
    >>> from evidently.calculations.stattests import chi_stat_test
    >>> options = DataDriftOptions(all_features_stattest=chi_stat_test)

    Using by name:

    >>> from evidently.options import DataDriftOptions
    >>> options = DataDriftOptions(all_features_stattest="chisquare")
"""
from typing import Tuple

import pandas as pd
from scipy.stats import chisquare

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest
from evidently.calculations.stattests.utils import get_unique_not_nan_values_list_from_series


def _chi_stat_test(
    reference_data: pd.Series, current_data: pd.Series, feature_type: str, threshold: float
) -> Tuple[float, bool]:
    keys = get_unique_not_nan_values_list_from_series(current_data=current_data, reference_data=reference_data)
    ref_feature_dict = {**dict.fromkeys(keys, 0), **dict(reference_data.value_counts())}
    current_feature_dict = {**dict.fromkeys(keys, 0), **dict(current_data.value_counts())}
    k_norm = current_data.shape[0] / reference_data.shape[0]
    f_exp = [ref_feature_dict[key] * k_norm for key in keys]
    f_obs = [current_feature_dict[key] for key in keys]
    p_value = chisquare(f_obs, f_exp)[1]
    return p_value, p_value < threshold


chi_stat_test = StatTest(
    name="chisquare", display_name="chi-square p_value", func=_chi_stat_test, allowed_feature_types=["cat"]
)

register_stattest(chi_stat_test)



import argparse
import json
import logging
import os
import sys
from dataclasses import dataclass
from typing import Any
from typing import Dict

import yaml

from evidently._config import TELEMETRY_ADDRESS
from evidently._config import TELEMETRY_ENABLED
from evidently.pipeline.column_mapping import ColumnMapping
from evidently.runner.dashboard_runner import DashboardRunner
from evidently.runner.dashboard_runner import DashboardRunnerOptions
from evidently.runner.loader import SamplingOptions
from evidently.runner.profile_runner import ProfileRunner
from evidently.runner.profile_runner import ProfileRunnerOptions
from evidently.runner.runner import DataOptions
from evidently.runner.runner import parse_options
from evidently.telemetry import TelemetrySender


@dataclass
class DataFormatOptions:
    header: bool
    separator: str
    date_column: str


@dataclass
class Sampling:
    reference: SamplingOptions
    current: SamplingOptions


@dataclass
class CalculateOptions:
    data_format: DataFormatOptions
    column_mapping: Dict[str, Any]
    sampling: Sampling


@dataclass
class DashboardOptions(CalculateOptions):
    dashboard_tabs: Dict[str, Dict[str, object]]


@dataclass
class ProfileOptions(CalculateOptions):
    profile_parts: Dict[str, Dict[str, str]]
    pretty_print: bool = False


def __get_not_none(src: Dict, key, default):
    return default if src.get(key, None) is None else src.get(key)


def __load_config_file(config_file: str):
    with open(config_file, encoding="utf-8") as f_config:
        if config_file.endswith(".yaml") or config_file.endswith(".yml"):
            opts_data = yaml.load(f_config, Loader=yaml.SafeLoader)
        elif config_file.endswith(".json"):
            opts_data = json.load(f_config)
        else:
            raise Exception(f"config .{config_file.split('.')[-1]} not supported")
    return opts_data


def calculate_dashboard(config: str, reference: str, current: str, output_path: str, report_name: str, **_kv):
    usage = dict(type="dashboard")

    opts_data = __load_config_file(config)

    sampling = __get_not_none(opts_data, "sampling", {})
    ref_sampling = __get_not_none(sampling, "reference", {})
    cur_sampling = __get_not_none(sampling, "current", {})

    usage["tabs"] = opts_data["dashboard_tabs"]
    usage["sampling"] = sampling
    opts = DashboardOptions(
        data_format=DataFormatOptions(**opts_data["data_format"]),
        column_mapping=opts_data["column_mapping"],
        dashboard_tabs=opts_data["dashboard_tabs"],
        sampling=Sampling(
            reference=SamplingOptions(**ref_sampling),
            current=SamplingOptions(**cur_sampling),
        ),
    )

    runner = DashboardRunner(
        DashboardRunnerOptions(
            reference_data_path=reference,
            reference_data_options=DataOptions(
                date_column=opts.data_format.date_column,
                separator=opts.data_format.separator,
                header=opts.data_format.header,
            ),
            reference_data_sampling=opts.sampling.reference,
            current_data_path=current,
            current_data_options=DataOptions(
                date_column=opts.data_format.date_column,
                separator=opts.data_format.separator,
                header=opts.data_format.header,
            ),
            current_data_sampling=opts.sampling.current,
            dashboard_tabs=opts.dashboard_tabs,
            options=parse_options(opts_data["options"]),
            column_mapping=ColumnMapping(**opts.column_mapping),
            output_path=os.path.join(output_path, report_name),
        )
    )
    runner.run()
    if TELEMETRY_ENABLED:
        sender = TelemetrySender(TELEMETRY_ADDRESS)
        sender.send(usage)


def calculate_profile(config: str, reference: str, current: str, output_path: str, report_name: str, **_kv):
    usage = dict(type="profile")

    opts_data = __load_config_file(config)

    sampling = __get_not_none(opts_data, "sampling", {})
    ref_sampling = __get_not_none(sampling, "reference", {})
    cur_sampling = __get_not_none(sampling, "current", {})
    usage["parts"] = opts_data["profile_sections"]
    usage["sampling"] = sampling
    opts = ProfileOptions(
        data_format=DataFormatOptions(**opts_data["data_format"]),
        column_mapping=opts_data["column_mapping"],
        profile_parts=opts_data["profile_sections"],
        pretty_print=opts_data["pretty_print"],
        sampling=Sampling(
            reference=SamplingOptions(**ref_sampling),
            current=SamplingOptions(**cur_sampling),
        ),
    )

    runner = ProfileRunner(
        ProfileRunnerOptions(
            reference_data_path=reference,
            reference_data_options=DataOptions(
                date_column=opts.data_format.date_column,
                separator=opts.data_format.separator,
                header=opts.data_format.header,
            ),
            reference_data_sampling=opts.sampling.reference,
            current_data_path=current,
            current_data_options=DataOptions(
                date_column=opts.data_format.date_column,
                separator=opts.data_format.separator,
                header=opts.data_format.header,
            ),
            current_data_sampling=opts.sampling.current,
            profile_parts=opts.profile_parts,
            column_mapping=ColumnMapping(**opts.column_mapping),
            options=parse_options(opts_data.get("options", None)),
            output_path=os.path.join(output_path, report_name),
            pretty_print=opts.pretty_print,
        )
    )
    runner.run()
    if TELEMETRY_ENABLED:
        sender = TelemetrySender(TELEMETRY_ADDRESS)
        sender.send(usage)


def help_handler(**_kv):
    parser.print_help()
    sys.exit(1)


def _add_default_parameters(configurable_parser, default_output_name: str):
    configurable_parser.add_argument("--reference", dest="reference", required=True, help="Path to reference data")
    configurable_parser.add_argument("--current", dest="current", help="Path to current data")
    configurable_parser.add_argument("--output_path", dest="output_path", required=True, help="Path to store report")
    configurable_parser.add_argument(
        "--report_name", dest="report_name", default=default_output_name, help="Report name"
    )
    configurable_parser.add_argument("--config", dest="config", required=True, help="Path to configuration")


logging.basicConfig(level=logging.INFO)

parser = argparse.ArgumentParser()

parsers = parser.add_subparsers()
parser.set_defaults(handler=help_handler)
calculate_parser = parsers.add_parser("calculate")
calc_subparsers = calculate_parser.add_subparsers()
profile_parser = calc_subparsers.add_parser("profile")
_add_default_parameters(profile_parser, "profile")
profile_parser.set_defaults(handler=calculate_profile)
dashboard_parser = calc_subparsers.add_parser("dashboard")
_add_default_parameters(dashboard_parser, "dashboard")
dashboard_parser.set_defaults(handler=calculate_dashboard)

parsed = parser.parse_args(sys.argv[1:])

parsed.handler(**parsed.__dict__)



from typing import List
from typing import Optional

from evidently.base_metric import InputData
from evidently.test_preset.test_preset import TestPreset
from evidently.tests import TestAllColumnsShareOfMissingValues
from evidently.tests import TestCatColumnsOutOfListValues
from evidently.tests import TestColumnsType
from evidently.tests import TestNumberOfColumns
from evidently.tests import TestNumberOfRows
from evidently.tests import TestNumColumnsMeanInNSigmas
from evidently.tests import TestNumColumnsOutOfRangeValues
from evidently.utils.data_operations import DatasetColumns


class DataStabilityTestPreset(TestPreset):
    """
    Data Stability tests.

    Contains tests:
    - `TestNumberOfRows`
    - `TestNumberOfColumns`
    - `TestColumnsType`
    - `TestAllColumnsShareOfMissingValues`
    - `TestNumColumnsOutOfRangeValues`
    - `TestCatColumnsOutOfListValues`
    - `TestNumColumnsMeanInNSigmas`
    """

    columns: Optional[List[str]]

    def __init__(
        self,
        columns: Optional[List[str]] = None,
    ):
        super().__init__()
        self.columns = columns

    def generate_tests(self, data: InputData, columns: DatasetColumns):
        return [
            TestNumberOfRows(),
            TestNumberOfColumns(),
            TestColumnsType(),
            TestAllColumnsShareOfMissingValues(columns=self.columns),
            TestNumColumnsOutOfRangeValues(columns=self.columns),
            TestCatColumnsOutOfListValues(columns=self.columns),
            TestNumColumnsMeanInNSigmas(columns=self.columns),
        ]



"""Anderson-Darling test of two samples.

Name: "anderson"

Import:

    >>> from evidently.calculations.stattests import anderson_darling_test

Properties:
- only for numerical features
- returns p-value

Example:
    Using by object:

    >>> from evidently.options import DataDriftOptions
    >>> from evidently.calculations.stattests import anderson_darling_test
    >>> options = DataDriftOptions(all_features_stattest=anderson_darling_test)

    Using by name:

    >>> from evidently.options import DataDriftOptions
    >>> options = DataDriftOptions(all_features_stattest="anderson")
"""
from typing import Tuple

import numpy as np
import pandas as pd
from scipy.stats import anderson_ksamp

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest


def _anderson_darling(
    reference_data: pd.Series,
    current_data: pd.Series,
    feature_type: str,
    threshold: float,
) -> Tuple[float, bool]:
    p_value = anderson_ksamp(np.array([reference_data, current_data]))[2]
    return p_value, p_value < threshold


anderson_darling_test = StatTest(
    name="anderson",
    display_name="Anderson-Darling",
    func=_anderson_darling,
    allowed_feature_types=["num"],
    default_threshold=0.1,
)

register_stattest(anderson_darling_test)



import abc
import logging
from dataclasses import dataclass
from enum import Enum
from typing import Generic
from typing import List
from typing import Optional
from typing import TypeVar
from typing import Union

import pandas as pd

from evidently.features.generated_features import GeneratedFeature
from evidently.pipeline.column_mapping import ColumnMapping
from evidently.utils.data_preprocessing import DataDefinition

TResult = TypeVar("TResult")


class ErrorResult:
    exception: BaseException

    def __init__(self, exception: BaseException):
        self.exception = exception


class DatasetType(Enum):
    MAIN = "main"
    ADDITIONAL = "additional"


@dataclass(eq=True, unsafe_hash=True)
class ColumnName:
    name: str
    dataset: DatasetType
    feature_class: Optional[GeneratedFeature]


def additional_feature(feature: GeneratedFeature, feature_name: str) -> ColumnName:
    return ColumnName(
        name=feature.__class__.__name__ + "." + feature_name,
        dataset=DatasetType.ADDITIONAL,
        feature_class=feature,
    )


@dataclass
class InputData:
    reference_data: Optional[pd.DataFrame]
    current_data: pd.DataFrame
    reference_additional_features: Optional[pd.DataFrame]
    current_additional_features: Optional[pd.DataFrame]
    column_mapping: ColumnMapping
    data_definition: DataDefinition

    @staticmethod
    def _get_by_column_name(dataset: pd.DataFrame, additional: pd.DataFrame, column: ColumnName) -> pd.Series:
        if column.dataset == DatasetType.MAIN:
            return dataset[column.name]
        if column.dataset == DatasetType.ADDITIONAL:
            return additional[column.name]
        raise ValueError("unknown column data")

    def get_current_column(self, column: Union[str, ColumnName]) -> pd.Series:
        if isinstance(column, str):
            _column = ColumnName(column, DatasetType.MAIN, None)
        else:
            _column = column
        return self._get_by_column_name(self.current_data, self.current_additional_features, _column)

    def get_reference_column(self, column: Union[str, ColumnName]) -> Optional[pd.Series]:
        if self.reference_data is None:
            return None
        if isinstance(column, str):
            _column = ColumnName(column, DatasetType.MAIN, None)
        else:
            _column = column
        if self.reference_additional_features is None and _column.dataset == DatasetType.ADDITIONAL:
            return None
        return self._get_by_column_name(self.reference_data, self.reference_additional_features, _column)


class Metric(Generic[TResult]):
    context = None

    def get_id(self) -> str:
        return self.__class__.__name__

    @abc.abstractmethod
    def calculate(self, data: InputData) -> TResult:
        raise NotImplementedError()

    def set_context(self, context):
        self.context = context

    def get_result(self) -> TResult:
        if self.context is None:
            raise ValueError("No context is set")
        result = self.context.metric_results.get(self, None)
        if isinstance(result, ErrorResult):
            raise result.exception
        if result is None:
            raise ValueError(f"No result found for metric {self} of type {type(self).__name__}")
        return result

    def get_parameters(self) -> Optional[tuple]:
        attributes = []
        for field, value in sorted(self.__dict__.items(), key=lambda x: x[0]):
            if field in ["context"]:
                continue
            if isinstance(value, list):
                attributes.append(tuple(value))
            else:
                attributes.append(value)
        params = tuple(attributes)
        try:
            hash(params)
        except TypeError:
            logging.warning(f"unhashable params for {type(self)}. Fallback to unique.")
            return None
        return params

    def required_features(self, data_definition: DataDefinition) -> List[GeneratedFeature]:
        required_features = []
        for field, value in sorted(self.__dict__.items(), key=lambda x: x[0]):
            if field in ["context"]:
                continue
            if issubclass(type(value), ColumnName) and value.feature_class is not None:
                required_features.append(value.feature_class)
        return required_features



from typing import Optional

import pandas as pd
import plotly.graph_objs as go

from evidently import ColumnMapping
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import AdditionalGraphInfo
from evidently.model.widget import BaseWidgetInfo


class TestSuiteWidget(Widget):
    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        fig = go.Figure()
        fig.add_trace(go.Bar(x=[1, 2, 3, 4, 5], y=[0.1, 0.2, 0.3, 0.4, 0.5]))

        fig_dict = fig.to_dict()
        return BaseWidgetInfo(
            title="",
            type="test_suite",
            size=2,
            params={
                "tests": [
                    {
                        "title": "Test Case #1",
                        "description": "Some long description of test case",
                        "state": "success",
                    },
                    {
                        "title": "Test Case #2",
                        "description": "Some long description of test case",
                        "state": "warning",
                    },
                    {
                        "title": "Test Case #3",
                        "description": "Some long description of test case",
                        "state": "fail",
                        "details": {
                            "parts": [
                                {
                                    "title": "Graph",
                                    "id": "test_case_3",
                                },
                                {
                                    "title": "Graph 2",
                                    "id": "test_case_3_2",
                                },
                            ]
                        },
                    },
                ]
            },
            additionalGraphs=[
                AdditionalGraphInfo(
                    "test_case_3",
                    {
                        "data": fig_dict["data"],
                        "layout": fig_dict["layout"],
                    },
                ),
                AdditionalGraphInfo(
                    "test_case_3_2",
                    {
                        "data": fig_dict["data"],
                        "layout": fig_dict["layout"],
                    },
                ),
            ],
        )

    def analyzers(self):
        return []



import dataclasses
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import pandas as pd
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import mean_absolute_percentage_error
from sklearn.metrics import mean_squared_error
from sklearn.metrics import r2_score

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.regression_performance import calculate_regression_performance
from evidently.metrics.utils import apply_func_to_binned_data
from evidently.metrics.utils import make_target_bins_for_reg_plots
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import CounterData
from evidently.renderers.html_widgets import counter
from evidently.renderers.html_widgets import header_text
from evidently.utils.data_operations import DatasetColumns
from evidently.utils.data_operations import process_columns
from evidently.utils.visualizations import make_hist_for_cat_plot
from evidently.utils.visualizations import make_hist_for_num_plot


@dataclass
class RegressionPerformanceMetricsResults:
    columns: DatasetColumns
    r2_score: float
    rmse: float
    rmse_default: float
    mean_error: float
    me_default_sigma: float
    me_hist_for_plot: Dict[str, Union[pd.Series, pd.DataFrame]]
    mean_abs_error: float
    mean_abs_error_default: float
    mean_abs_perc_error: float
    mean_abs_perc_error_default: float
    abs_error_max: float
    abs_error_max_default: float
    error_std: float
    abs_error_std: float
    abs_perc_error_std: float
    error_normality: dict
    underperformance: dict
    hist_for_plot: Dict[str, pd.Series]
    vals_for_plots: Dict[str, Dict[str, pd.Series]]
    error_bias: Optional[dict] = None
    mean_error_ref: Optional[float] = None
    mean_abs_error_ref: Optional[float] = None
    mean_abs_perc_error_ref: Optional[float] = None
    rmse_ref: Optional[float] = None
    r2_score_ref: Optional[float] = None
    abs_error_max_ref: Optional[float] = None
    underperformance_ref: Optional[dict] = None


class RegressionPerformanceMetrics(Metric[RegressionPerformanceMetricsResults]):
    def get_parameters(self) -> tuple:
        return ()

    def calculate(self, data: InputData) -> RegressionPerformanceMetricsResults:
        columns = process_columns(data.current_data, data.column_mapping)

        current_metrics = calculate_regression_performance(
            dataset=data.current_data, columns=columns, error_bias_prefix="current_"
        )
        error_bias = current_metrics.error_bias
        reference_metrics = None

        if data.reference_data is not None:
            ref_columns = process_columns(data.reference_data, data.column_mapping)
            reference_metrics = calculate_regression_performance(
                dataset=data.reference_data, columns=ref_columns, error_bias_prefix="ref_"
            )

            if reference_metrics is not None and reference_metrics.error_bias:
                for feature_name, current_bias in reference_metrics.error_bias.items():
                    if feature_name in error_bias:
                        error_bias[feature_name].update(current_bias)

                    else:
                        error_bias[feature_name] = current_bias

        r2_score_value = r2_score(
            y_true=data.current_data[data.column_mapping.target],
            y_pred=data.current_data[data.column_mapping.prediction],
        )
        rmse_score_value = mean_squared_error(
            y_true=data.current_data[data.column_mapping.target],
            y_pred=data.current_data[data.column_mapping.prediction],
        )

        # mae default values
        dummy_preds = data.current_data[data.column_mapping.target].median()
        mean_abs_error_default = mean_absolute_error(
            y_true=data.current_data[data.column_mapping.target], y_pred=[dummy_preds] * data.current_data.shape[0]
        )
        # rmse default values
        rmse_ref = None
        if data.reference_data is not None:
            rmse_ref = mean_squared_error(
                y_true=data.reference_data[data.column_mapping.target],
                y_pred=data.reference_data[data.column_mapping.prediction],
            )
        dummy_preds = data.current_data[data.column_mapping.target].mean()
        rmse_default = mean_squared_error(
            y_true=data.current_data[data.column_mapping.target], y_pred=[dummy_preds] * data.current_data.shape[0]
        )
        # mape default values
        # optimal constant for mape
        s = data.current_data[data.column_mapping.target]
        inv_y = 1 / s[s != 0].values
        w = inv_y / sum(inv_y)
        idxs = np.argsort(w)
        sorted_w = w[idxs]
        sorted_w_cumsum = np.cumsum(sorted_w)
        idx = np.where(sorted_w_cumsum > 0.5)[0][0]
        pos = idxs[idx]
        dummy_preds = s[s != 0].values[pos]

        mean_abs_perc_error_default = (
            mean_absolute_percentage_error(
                y_true=data.current_data[data.column_mapping.target],
                y_pred=[dummy_preds] * data.current_data.shape[0],
            )
            * 100
        )
        #  r2_score default values
        r2_score_ref = None
        if data.reference_data is not None:
            r2_score_ref = r2_score(
                y_true=data.reference_data[data.column_mapping.target],
                y_pred=data.reference_data[data.column_mapping.prediction],
            )
        # max error default values
        abs_error_max_ref = None
        mean_error_ref = None

        if reference_metrics is not None:
            abs_error_max_ref = reference_metrics.abs_error_max
            mean_error_ref = reference_metrics.mean_error

        y_true = data.current_data[data.column_mapping.target]
        y_pred = data.current_data[data.column_mapping.prediction]
        abs_error_max_default = np.abs(y_true - y_true.median()).max()

        #  me default values
        me_default_sigma = (y_pred - y_true).std()

        # visualisation

        df_target_binned = make_target_bins_for_reg_plots(
            data.current_data, data.column_mapping.target, data.column_mapping.prediction, data.reference_data
        )
        curr_target_bins = df_target_binned.loc[df_target_binned.data == "curr", "target_binned"]
        ref_target_bins = None
        if data.reference_data is not None:
            ref_target_bins = df_target_binned.loc[df_target_binned.data == "ref", "target_binned"]
        hist_for_plot = make_hist_for_cat_plot(curr_target_bins, ref_target_bins)

        vals_for_plots = {}

        if data.reference_data is not None:
            is_ref_data = True

        else:
            is_ref_data = False

        for name, func in zip(
            ["r2_score", "rmse", "mean_abs_error", "mean_abs_perc_error"],
            [r2_score, mean_squared_error, mean_absolute_error, mean_absolute_percentage_error],
        ):
            vals_for_plots[name] = apply_func_to_binned_data(
                df_target_binned, func, data.column_mapping.target, data.column_mapping.prediction, is_ref_data
            )

        # me plot
        err_curr = data.current_data[data.column_mapping.prediction] - data.current_data[data.column_mapping.target]
        err_ref = None

        if is_ref_data:
            err_ref = (
                data.reference_data[data.column_mapping.prediction] - data.reference_data[data.column_mapping.target]
            )
        me_hist_for_plot = make_hist_for_num_plot(err_curr, err_ref)

        if r2_score_ref is not None:
            r2_score_ref = float(r2_score_ref)

        if rmse_ref is not None:
            rmse_ref = float(rmse_ref)

        underperformance_ref = None

        if reference_metrics is not None:
            underperformance_ref = reference_metrics.underperformance

        return RegressionPerformanceMetricsResults(
            columns=columns,
            r2_score=r2_score_value,
            rmse=rmse_score_value,
            rmse_default=rmse_default,
            mean_error=current_metrics.mean_error,
            mean_error_ref=mean_error_ref,
            me_default_sigma=me_default_sigma,
            me_hist_for_plot=me_hist_for_plot,
            mean_abs_error=current_metrics.mean_abs_error,
            mean_abs_error_default=mean_abs_error_default,
            mean_abs_perc_error=current_metrics.mean_abs_perc_error,
            mean_abs_perc_error_default=mean_abs_perc_error_default,
            abs_error_max=current_metrics.abs_error_max,
            abs_error_max_default=abs_error_max_default,
            error_std=current_metrics.error_std,
            abs_error_std=current_metrics.abs_error_std,
            abs_perc_error_std=current_metrics.abs_perc_error_std,
            error_normality=current_metrics.error_normality,
            underperformance=current_metrics.underperformance,
            underperformance_ref=underperformance_ref,
            hist_for_plot=hist_for_plot,
            vals_for_plots=vals_for_plots,
            error_bias=error_bias,
            mean_abs_error_ref=reference_metrics.mean_abs_error if reference_metrics is not None else None,
            mean_abs_perc_error_ref=reference_metrics.mean_abs_perc_error if reference_metrics is not None else None,
            rmse_ref=rmse_ref,
            r2_score_ref=r2_score_ref,
            abs_error_max_ref=abs_error_max_ref,
        )


@default_renderer(wrap_type=RegressionPerformanceMetrics)
class RegressionPerformanceMetricsRenderer(MetricRenderer):
    def render_json(self, obj: RegressionPerformanceMetrics) -> dict:
        result = dataclasses.asdict(obj.get_result())
        # remove values with DataFrames or Series
        result.pop("hist_for_plot")
        result.pop("vals_for_plots")
        result.pop("me_hist_for_plot")
        return result

    @staticmethod
    def _get_underperformance_tails(dataset_name: str, underperformance: dict) -> BaseWidgetInfo:
        return counter(
            title=f"{dataset_name.capitalize()}: Mean Error per Group (+/- std)",
            counters=[
                CounterData.float("Majority(90%)", underperformance["majority"]["mean_error"], 2),
                CounterData.float("Underestimation(5%)", underperformance["underestimation"]["mean_error"], 2),
                CounterData.float("Overestimation(5%)", underperformance["overestimation"]["mean_error"], 2),
            ],
        )

    def render_html(self, obj: RegressionPerformanceMetrics) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        target_name = metric_result.columns.utility_columns.target
        result = [
            header_text(label=f"Regression Model Performance. Target: '{target_name}’"),
            counter(
                title="Current: Regression Performance Metrics",
                counters=[
                    CounterData.float("Mean error", metric_result.mean_error, 3),
                    CounterData.float("MAE", metric_result.mean_abs_error, 3),
                    CounterData.float("MAPE", metric_result.mean_abs_perc_error, 3),
                    CounterData.float("RMSE", metric_result.rmse, 3),
                    CounterData.float("r2 score", metric_result.r2_score, 3),
                ],
            ),
        ]
        if (
            metric_result.mean_error_ref is not None
            and metric_result.mean_abs_error_ref is not None
            and metric_result.mean_abs_perc_error_ref is not None
            and metric_result.rmse_ref is not None
            and metric_result.r2_score_ref is not None
        ):
            result.append(
                counter(
                    title="Reference: Regression Performance Metrics",
                    counters=[
                        CounterData.float("Mean error", metric_result.mean_error_ref, 3),
                        CounterData.float("MAE", metric_result.mean_abs_error_ref, 3),
                        CounterData.float("MAPE", metric_result.mean_abs_perc_error_ref, 3),
                        CounterData.float("RMSE", metric_result.rmse_ref, 3),
                        CounterData.float("r2 score", metric_result.r2_score_ref, 3),
                    ],
                ),
            )

        return result



import pandas as pd
from plotly.graph_objs import Figure


def fig_to_json(figure: Figure) -> dict:
    result = figure.to_plotly_json()
    result["layout"].pop("template", None)
    return result


class CutQuantileTransformer:
    def __init__(self, side: str, q: float):
        self.side = side
        self.q = q
        self.q_val_left = None
        self.q_val_right = None

    def fit(self, s: pd.Series):
        if self.side == "left":
            self.q_val_left = s.quantile(self.q)
        elif self.side == "right":
            self.q_val_right = s.quantile(self.q)
        else:
            self.q_val_left = s.quantile(self.q)
            self.q_val_right = s.quantile(1 - self.q)

    def transform(self, s: pd.Series):
        if self.side == "left":
            return s[s >= self.q_val_left]
        elif self.side == "right":
            return s[s <= self.q_val_right]
        else:
            return s[s.between(self.q_val_left, self.q_val_right)]

    def transform_df(self, df: pd.DataFrame, feature: str):
        if self.side == "left":
            return df[df[feature] >= self.q_val_left]
        elif self.side == "right":
            return df[df[feature] <= self.q_val_right]
        else:
            return df[df[feature].between(self.q_val_left, self.q_val_right)]



import dataclasses
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional

import numpy as np
import pandas as pd
from plotly import figure_factory as ff

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import GraphData
from evidently.renderers.html_widgets import WidgetSize
from evidently.renderers.html_widgets import plotly_graph_tabs
from evidently.utils.data_operations import process_columns


@dataclasses.dataclass
class ClassificationProbDistributionResults:
    current_distribution: Optional[Dict[str, list]]
    reference_distribution: Optional[Dict[str, list]]


class ClassificationProbDistribution(Metric[ClassificationProbDistributionResults]):
    @staticmethod
    def get_distribution(dataset: pd.DataFrame, target_name: str, prediction_labels: Iterable) -> Dict[str, list]:
        result = {}
        dataset.replace([np.inf, -np.inf], np.nan, inplace=True)

        for label in prediction_labels:
            result[label] = [
                dataset[dataset[target_name] == label][label],
                dataset[dataset[target_name] != label][label],
            ]

        return result

    def calculate(self, data: InputData) -> ClassificationProbDistributionResults:
        columns = process_columns(data.current_data, data.column_mapping)
        prediction = columns.utility_columns.prediction
        target = columns.utility_columns.target

        if target is None:
            raise ValueError("Target column should be present")

        if not isinstance(prediction, Iterable) or isinstance(prediction, str):
            current_distribution = None
            reference_distribution = None

        else:
            current_data_copy = data.current_data.copy()
            current_distribution = self.get_distribution(current_data_copy, target, prediction)

            if data.reference_data is not None:
                reference_data_copy = data.reference_data.copy()
                reference_distribution = self.get_distribution(reference_data_copy, target, prediction)

            else:
                reference_distribution = None

        return ClassificationProbDistributionResults(
            current_distribution=current_distribution,
            reference_distribution=reference_distribution,
        )


@default_renderer(wrap_type=ClassificationProbDistribution)
class ClassificationProbDistributionRenderer(MetricRenderer):
    def render_json(self, obj: ClassificationProbDistribution) -> dict:
        return {}

    def _plot(self, distribution: Dict[str, list]):
        # plot distributions
        graphs = []

        for label in distribution:
            pred_distr = ff.create_distplot(
                distribution[label],
                [str(label), "other"],
                colors=[self.color_options.primary_color, self.color_options.secondary_color],
                bin_size=0.05,
                show_curve=False,
                show_rug=True,
            )

            pred_distr.update_layout(
                xaxis_title="Probability",
                yaxis_title="Share",
                legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1),
            )
            pred_distr_json = pred_distr.to_plotly_json()
            graphs.append(
                {
                    "title": str(label),
                    "data": pred_distr_json["data"],
                    "layout": pred_distr_json["layout"],
                }
            )
        return graphs

    def render_html(self, obj: ClassificationProbDistribution) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        reference_distribution = metric_result.reference_distribution
        current_distribution = metric_result.current_distribution
        result = []
        size = WidgetSize.FULL

        if reference_distribution is not None:
            size = WidgetSize.HALF

        if current_distribution is not None:
            result.append(
                plotly_graph_tabs(
                    title="Current: Probability Distribution",
                    size=size,
                    figures=[
                        GraphData(graph["title"], graph["data"], graph["layout"])
                        for graph in self._plot(current_distribution)
                    ],
                )
            )

        if reference_distribution is not None:
            result.append(
                plotly_graph_tabs(
                    title="Reference: Probability Distribution",
                    size=size,
                    figures=[
                        GraphData(graph["title"], graph["data"], graph["layout"])
                        for graph in self._plot(reference_distribution)
                    ],
                )
            )

        return result



from typing import Tuple

import numpy as np
import pandas as pd


def make_hist_df(hist: Tuple[np.array, np.array]) -> pd.DataFrame:
    hist_df = pd.DataFrame(
        np.array(
            [
                hist[1][:-1],
                hist[0],
                [f"{x[0]}-{x[1]}" for x in zip(hist[1][:-1], hist[1][1:])],
            ]
        ).T,
        columns=["x", "count", "range"],
    )

    hist_df["x"] = hist_df["x"].astype(float)
    hist_df["count"] = hist_df["count"].astype(int)
    return hist_df


def make_hist_for_num_plot(curr: pd.Series, ref: pd.Series = None):
    result = {}
    if ref is not None:
        ref = ref.dropna()
    bins = np.histogram_bin_edges(pd.concat([curr.dropna(), ref]), bins="doane")
    curr_hist = np.histogram(curr, bins=bins)
    result["current"] = make_hist_df(curr_hist)
    if ref is not None:
        ref_hist = np.histogram(ref, bins=bins)
        result["reference"] = make_hist_df(ref_hist)
    return result


def make_hist_for_cat_plot(curr: pd.Series, ref: pd.Series = None, normalize: bool = False, dropna=False):
    result = {}
    hist_df = curr.astype(str).value_counts(normalize=normalize, dropna=dropna).reset_index()
    hist_df.columns = ["x", "count"]
    result["current"] = hist_df
    if ref is not None:
        hist_df = ref.astype(str).value_counts(normalize=normalize, dropna=dropna).reset_index()
        hist_df.columns = ["x", "count"]
        result["reference"] = hist_df
    return result



from typing import Generator
from typing import List
from typing import Type

from evidently.analyzers.base_analyzer import Analyzer
from evidently.analyzers.regression_performance_analyzer import (
    RegressionPerformanceAnalyzer,
)
from evidently.analyzers.regression_performance_analyzer import (
    RegressionPerformanceMetrics,
)
from evidently.model_monitoring.monitoring import MetricsType
from evidently.model_monitoring.monitoring import ModelMonitor
from evidently.model_monitoring.monitoring import ModelMonitoringMetric


class RegressionPerformanceMonitorMetrics:
    _tag = "regression_performance"
    quality = ModelMonitoringMetric(f"{_tag}:quality", ["dataset", "metric"])
    normality = ModelMonitoringMetric(f"{_tag}:error_normality", ["dataset", "metric"])
    underperformance = ModelMonitoringMetric(
        f"{_tag}:underperformance", ["dataset", "metric", "type"]
    )
    feature_error_bias = ModelMonitoringMetric(
        f"{_tag}:feature_error_bias", ["feature", "feature_type", "metric"]
    )


class RegressionPerformanceMonitor(ModelMonitor):
    def monitor_id(self) -> str:
        return "regression_performance"

    def analyzers(self) -> List[Type[Analyzer]]:
        return [RegressionPerformanceAnalyzer]

    def metrics(self, analyzer_results) -> Generator[MetricsType, None, None]:
        results = RegressionPerformanceAnalyzer.get_results(analyzer_results)

        if results.reference_metrics is not None:
            for metric in self._yield_quality(results.reference_metrics, "reference"):
                yield metric

            for metric in self._yield_error_normality(
                results.reference_metrics.error_normality, "reference"
            ):
                yield metric

            for metric in self._yield_underperformance(
                results.reference_metrics.underperformance, "reference"
            ):
                yield metric

        if results.current_metrics is not None:
            for metric in self._yield_quality(results.current_metrics, "current"):
                yield metric

            for metric in self._yield_error_normality(
                results.current_metrics.error_normality, "current"
            ):
                yield metric

            for metric in self._yield_underperformance(
                results.current_metrics.underperformance, "current"
            ):
                yield metric

        fields = [
            "ref_majority",
            "ref_under",
            "ref_over",
            "ref_range",
            "current_majority",
            "current_under",
            "current_over",
            "current_range",
        ]

        if results.error_bias is not None:
            for feature in results.columns.num_feature_names:
                for field in fields:
                    yield RegressionPerformanceMonitorMetrics.feature_error_bias.create(
                        results.error_bias[feature][field],
                        dict(feature=feature, feature_type="num", metric=field),
                    )

        if results.error_bias is not None:
            for feature in results.columns.cat_feature_names:
                for field in fields:
                    yield RegressionPerformanceMonitorMetrics.feature_error_bias.create(
                        results.error_bias[feature][field],
                        dict(feature=feature, feature_type="cat", metric=field),
                    )

    @staticmethod
    def _yield_quality(
        metrics: RegressionPerformanceMetrics, dataset: str
    ) -> Generator[MetricsType, None, None]:
        yield RegressionPerformanceMonitorMetrics.quality.create(
            metrics.mean_error, dict(dataset=dataset, metric="mean_error")
        )
        yield RegressionPerformanceMonitorMetrics.quality.create(
            metrics.mean_abs_error, dict(dataset=dataset, metric="mean_abs_error")
        )
        yield RegressionPerformanceMonitorMetrics.quality.create(
            metrics.mean_abs_perc_error,
            dict(dataset=dataset, metric="mean_abs_perc_error"),
        )
        yield RegressionPerformanceMonitorMetrics.quality.create(
            metrics.error_std, dict(dataset=dataset, metric="error_std")
        )
        yield RegressionPerformanceMonitorMetrics.quality.create(
            metrics.abs_error_std, dict(dataset=dataset, metric="abs_error_std")
        )
        yield RegressionPerformanceMonitorMetrics.quality.create(
            metrics.abs_perc_error_std,
            dict(dataset=dataset, metric="abs_perc_error_std"),
        )

    @staticmethod
    def _yield_error_normality(
        normality_data, dataset
    ) -> Generator[MetricsType, None, None]:
        metric_labels = ["slope", "intercept", "r"]
        for label in metric_labels:
            yield RegressionPerformanceMonitorMetrics.normality.create(
                normality_data[label], dict(dataset=dataset, metric=label)
            )

    @staticmethod
    def _yield_underperformance(
        underperformance_data, dataset
    ) -> Generator[MetricsType, None, None]:
        type_labels = ["majority", "underestimation", "overestimation"]
        metric_labels = ["mean_error", "std_error"]
        for type_label in type_labels:
            for metric_label in metric_labels:
                yield RegressionPerformanceMonitorMetrics.underperformance.create(
                    underperformance_data[type_label][metric_label],
                    dict(dataset=dataset, metric=metric_label, type=type_label),
                )



import dataclasses
from typing import List
from typing import Optional

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import CounterData
from evidently.renderers.html_widgets import counter
from evidently.renderers.html_widgets import header_text
from evidently.utils.data_operations import process_columns


@dataclasses.dataclass
class ConflictTargetMetricResults:
    number_not_stable_target: int
    share_not_stable_target: float
    number_not_stable_target_ref: Optional[int] = None
    share_not_stable_target_ref: Optional[float] = None


class ConflictTargetMetric(Metric[ConflictTargetMetricResults]):
    def calculate(self, data: InputData) -> ConflictTargetMetricResults:
        dataset_columns = process_columns(data.current_data, data.column_mapping)
        target_name = dataset_columns.utility_columns.target
        if target_name is None:
            raise ValueError("The column 'target' should be presented")
        columns = dataset_columns.get_all_features_list()
        if len(columns) == 0:
            raise ValueError("Target conflict is not defined. No features provided")
        duplicates = data.current_data[data.current_data.duplicated(subset=columns, keep=False)]
        number_not_stable_target = duplicates.drop(
            data.current_data[data.current_data.duplicated(subset=columns + [target_name], keep=False)].index
        ).shape[0]
        share_not_stable_target = round(number_not_stable_target / data.current_data.shape[0], 3)
        # reference
        number_not_stable_target_ref = None
        share_not_stable_target_ref = None
        if data.reference_data is not None:
            duplicates_ref = data.reference_data[data.reference_data.duplicated(subset=columns, keep=False)]
            number_not_stable_target_ref = duplicates_ref.drop(
                data.reference_data[data.reference_data.duplicated(subset=columns + [target_name], keep=False)].index
            ).shape[0]
            share_not_stable_target_ref = round(number_not_stable_target_ref / data.reference_data.shape[0], 3)
        return ConflictTargetMetricResults(
            number_not_stable_target=number_not_stable_target,
            share_not_stable_target=share_not_stable_target,
            number_not_stable_target_ref=number_not_stable_target_ref,
            share_not_stable_target_ref=share_not_stable_target_ref,
        )


@default_renderer(wrap_type=ConflictTargetMetric)
class ConflictTargetMetricRenderer(MetricRenderer):
    def render_json(self, obj: ConflictTargetMetric) -> dict:
        return dataclasses.asdict(obj.get_result())

    def render_html(self, obj: ConflictTargetMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        counters = [
            CounterData(
                "number of conflicts (current)",
                self._get_string(metric_result.number_not_stable_target, metric_result.share_not_stable_target),
            )
        ]
        if (
            metric_result.number_not_stable_target_ref is not None
            and metric_result.share_not_stable_target_ref is not None
        ):
            counters.append(
                CounterData(
                    "number of conflicts (reference)",
                    self._get_string(
                        metric_result.number_not_stable_target_ref, metric_result.share_not_stable_target_ref
                    ),
                )
            )
        result = [
            header_text(label="Conflicts in Target"),
            counter(counters=counters),
        ]
        return result

    @staticmethod
    def _get_string(number: int, ratio: float) -> str:
        return f"{number} ({ratio * 100}%)"



#!/usr/bin/env python
# coding: utf-8
# pylint: disable=invalid-name
import dataclasses
import uuid
from dataclasses import dataclass
from enum import Enum
from typing import Any
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union


@dataclass()
class TriggeredAlertStats:
    period: int
    last_24h: int


@dataclass
class AlertStats:
    """
    Attributes:
        active: Number of active alerts.
        eggs: An integer count of the eggs we have laid.
    """

    active: int
    triggered: TriggeredAlertStats


@dataclass
class Insight:
    """
    Attributes:
        title: Insight title
        severity: Severity level for insight information (one of 'info', 'warning', 'error', 'success')
        text: Insidght information
    """

    title: str
    severity: str
    text: str


@dataclass
class Alert:
    value: Union[str, int, float]
    state: str
    text: str
    longText: str


@dataclass
class AdditionalGraphInfo:
    id: str
    params: Any


@dataclass
class PlotlyGraphInfo:
    data: Any
    layout: Any
    id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))


class WidgetType(Enum):
    COUNTER = "counter"
    TABLE = "table"
    BIG_TABLE = "big_table"
    BIG_GRAPH = "big_graph"
    RICH_DATA = "rich_data"
    TABBED_GRAPH = "tabbed_graph"
    TABS = "tabs"


@dataclass
class BaseWidgetInfo:
    type: str
    title: str
    size: int
    id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))
    details: str = ""
    alertsPosition: Optional[str] = None
    alertStats: Optional[AlertStats] = None
    params: Any = None
    insights: Iterable[Insight] = ()
    additionalGraphs: Iterable[Union[AdditionalGraphInfo, "BaseWidgetInfo", PlotlyGraphInfo]] = ()
    alerts: Iterable[Alert] = ()
    tabs: Iterable["TabInfo"] = ()
    widgets: Iterable["BaseWidgetInfo"] = ()
    pageSize: int = 5

    def get_additional_graphs(self) -> List[Union[AdditionalGraphInfo, PlotlyGraphInfo, "BaseWidgetInfo"]]:
        return list(self.additionalGraphs) + [
            graph for widget in self.widgets for graph in widget.get_additional_graphs()
        ]


@dataclass
class TabInfo:
    id: str
    title: str
    widget: BaseWidgetInfo



import warnings

import evidently.dashboard.tabs
from evidently.dashboard.tabs import *

__path__ = evidently.dashboard.tabs.__path__  # type: ignore

warnings.warn(
    "'import evidently.tabs' is deprecated, use 'import evidently.dashboard.tabs'"
)



#!/usr/bin/env python
# coding: utf-8

import json
from typing import Optional

import pandas as pd
import plotly.express as px

from evidently import ColumnMapping
from evidently.analyzers.classification_performance_analyzer import (
    ClassificationPerformanceAnalyzer,
)
from evidently.dashboard.widgets.utils import CutQuantileTransformer
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import AdditionalGraphInfo
from evidently.model.widget import BaseWidgetInfo
from evidently.options import QualityMetricsOptions


class ClassConfusionBasedFeatureDistrTable(Widget):
    def analyzers(self):
        return [ClassificationPerformanceAnalyzer]

    @staticmethod
    def _get_rows_per_page(features_count: int) -> int:
        """How mane rows of features we want to show on one page. If we have many features - show 10 on one page"""
        return min(features_count, 10)

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:

        results = ClassificationPerformanceAnalyzer.get_results(analyzers_results)
        quality_metrics_options = self.options_provider.get(QualityMetricsOptions)
        cut_quantile = quality_metrics_options.cut_quantile
        target_name = results.columns.utility_columns.target
        prediction_name = results.columns.utility_columns.prediction
        rows_per_page = self._get_rows_per_page(results.columns.get_features_len())

        if target_name is None or results.columns.utility_columns.prediction is None:
            raise ValueError(
                f"Widget {self.title} requires 'target' and 'prediction' columns."
            )

        if current_data is not None:
            additional_graphs_data = []
            params_data = []

            for feature_name in results.columns.get_all_features_list(
                cat_before_num=False
            ):
                # add data for table in params
                labels = sorted(set(reference_data[target_name]))

                params_data.append(
                    {
                        "details": {
                            "parts": [
                                {"title": "All", "id": "All" + "_" + str(feature_name)}
                            ]
                            + [
                                {
                                    "title": str(label),
                                    "id": feature_name + "_" + str(label),
                                }
                                for label in labels
                            ],
                            "insights": [],
                        },
                        "f1": feature_name,
                    }
                )

                # create confusion based plots
                reference_data["dataset"] = "Reference"
                current_data["dataset"] = "Current"
                quantile = quality_metrics_options.get_cut_quantile(feature_name)
                if cut_quantile and quantile is not None:
                    side, q = quantile
                    cqt = CutQuantileTransformer(side=side, q=q)
                    cqt.fit(reference_data[feature_name])
                    reference_data_to_plot = cqt.transform_df(
                        reference_data, feature_name
                    )
                    current_data_to_plot = cqt.transform_df(current_data, feature_name)
                else:
                    reference_data_to_plot = reference_data
                    current_data_to_plot = current_data
                merged_data = pd.concat([reference_data_to_plot, current_data_to_plot])

                fig = px.histogram(
                    merged_data,
                    x=feature_name,
                    color=target_name,
                    facet_col="dataset",
                    histnorm="",
                    barmode="overlay",
                    category_orders={"dataset": ["Reference", "Current"]},
                )

                fig_json = json.loads(fig.to_json())

                # write plot data in table as additional data
                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        "All" + "_" + str(feature_name),
                        {"data": fig_json["data"], "layout": fig_json["layout"]},
                    )
                )

                for label in labels:

                    def confusion_func(row, label=label):
                        return _confusion(row, target_name, prediction_name, label)

                    merged_data["Confusion"] = merged_data.apply(confusion_func, axis=1)

                    fig = px.histogram(
                        merged_data,
                        x=feature_name,
                        color="Confusion",
                        facet_col="dataset",
                        histnorm="",
                        barmode="overlay",
                        category_orders={
                            "dataset": ["Reference", "Current"],
                            "Confusion": ["TP", "TN", "FP", "FN"],
                        },
                    )
                    fig_json = json.loads(fig.to_json())

                    # write plot data in table as additional data
                    additional_graphs_data.append(
                        AdditionalGraphInfo(
                            feature_name + "_" + str(label),
                            {"data": fig_json["data"], "layout": fig_json["layout"]},
                        )
                    )

            return BaseWidgetInfo(
                title=self.title,
                type="big_table",
                size=2,
                params={
                    "rowsPerPage": rows_per_page,
                    "columns": [{"title": "Feature", "field": "f1"}],
                    "data": params_data,
                },
                additionalGraphs=additional_graphs_data,
            )
        # if no current data is set.
        additional_graphs_data = []
        params_data = []

        for feature_name in results.columns.get_all_features_list(cat_before_num=False):
            # add data for table in params
            labels = sorted(set(reference_data[target_name]))

            params_data.append(
                {
                    "details": {
                        "parts": [
                            {"title": "All", "id": "All" + "_" + str(feature_name)}
                        ]
                        + [
                            {"title": str(label), "id": feature_name + "_" + str(label)}
                            for label in labels
                        ],
                        "insights": [],
                    },
                    "f1": feature_name,
                }
            )

            # create confusion based plots
            quantile = quality_metrics_options.get_cut_quantile(feature_name)
            if cut_quantile and quantile is not None:
                side, q = quantile
                cqt = CutQuantileTransformer(side=side, q=q)
                cqt.fit(reference_data[feature_name])
                reference_data_to_plot = cqt.transform_df(reference_data, feature_name)
            else:
                reference_data_to_plot = reference_data

            fig = px.histogram(
                reference_data_to_plot,
                x=feature_name,
                color=target_name,
                histnorm="",
                barmode="overlay",
            )

            fig_json = json.loads(fig.to_json())

            # write plot data in table as additional data
            additional_graphs_data.append(
                AdditionalGraphInfo(
                    "All" + "_" + str(feature_name),
                    {"data": fig_json["data"], "layout": fig_json["layout"]},
                )
            )

            for label in labels:

                def _confusion_func(row, label=label):
                    return _confusion(row, target_name, prediction_name, label)

                reference_data["Confusion"] = reference_data.apply(
                    _confusion_func, axis=1
                )

                fig = px.histogram(
                    reference_data_to_plot,
                    x=feature_name,
                    color="Confusion",
                    histnorm="",
                    barmode="overlay",
                    category_orders={"Confusion": ["TP", "TN", "FP", "FN"]},
                )

                fig_json = json.loads(fig.to_json())

                # write plot data in table as additional data
                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_" + str(label),
                        {"data": fig_json["data"], "layout": fig_json["layout"]},
                    )
                )

        return BaseWidgetInfo(
            title=self.title,
            type="big_table",
            size=2,
            params={
                "rowsPerPage": rows_per_page,
                "columns": [{"title": "Feature", "field": "f1"}],
                "data": params_data,
            },
            additionalGraphs=additional_graphs_data,
        )


def _confusion(row, target_column, prediction_column, label):
    if row[target_column] == label and row[prediction_column] == label:
        return "TP"
    if row[target_column] != label and row[prediction_column] == label:
        return "FP"
    if row[target_column] == label and row[prediction_column] != label:
        return "FN"
    return "TN"  # last option



#!/usr/bin/env python
# coding: utf-8
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

import pandas

from evidently.analyzers.base_analyzer import Analyzer
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo
from evidently.pipeline.column_mapping import ColumnMapping
from evidently.pipeline.stage import PipelineStage

VerboseLevel = int


class Verbose:
    ALWAYS = -1
    SHORT = 0
    FULL = 1


class Tab(PipelineStage):
    widgets: List[Tuple[Widget, VerboseLevel]]
    _widgets: List[Widget]
    _widget_results: List[Optional[BaseWidgetInfo]]

    def __init__(
        self,
        verbose_level: VerboseLevel = None,
        include_widgets: List[Union[str, Widget]] = None,
    ):
        super().__init__()
        if verbose_level is None:
            verbose_level = Verbose.FULL
        self._widgets = []
        self._widget_results = []
        self.details_level = verbose_level
        predefined_widgets = {widget[0].title: widget[0] for widget in self.widgets}

        if include_widgets is not None:
            for widget in include_widgets:
                if isinstance(widget, str):
                    self._widgets.append(predefined_widgets[widget])
                elif isinstance(widget, Widget):
                    self._widgets.append(widget)
                else:
                    raise ValueError(f"Unexpected value: {widget}")
        else:
            for _widget in self.widgets:
                if _widget[1] > verbose_level:
                    continue
                self._widgets.append(_widget[0])
        for _widget_instance in self._widgets:
            for analyzer in _widget_instance.analyzers():
                self.add_analyzer(analyzer)

    def calculate(
        self,
        reference_data: pandas.DataFrame,
        current_data: pandas.DataFrame,
        column_mapping: ColumnMapping,
        analyzers_results: Dict[Type[Analyzer], object],
    ):
        self._widget_results.clear()
        for widget in self._widgets:
            widget.options_provider = self.options_provider
            self._widget_results.append(
                widget.calculate(
                    reference_data, current_data, column_mapping, analyzers_results
                )
            )

    def info(self) -> List[Optional[BaseWidgetInfo]]:
        return self._widget_results

    @classmethod
    def list_widgets(cls):
        return [widget[0].title for widget in cls.widgets]



import re

import pandas as pd
from nltk.stem.wordnet import WordNetLemmatizer

from evidently.features.generated_features import GeneratedFeature
from evidently.utils.data_preprocessing import DataDefinition


class TriggerWordsPresent(GeneratedFeature):
    def __init__(self, column_name: str, words_list=(), lemmatisize=True):
        self.lem = WordNetLemmatizer()
        self.column_name = column_name
        self.words_list = words_list
        self.lemmatisize = lemmatisize

    def generate_feature(self, data: pd.DataFrame, data_definition: DataDefinition) -> pd.DataFrame:
        def listed_words_present(s, words_list=(), lemmatisize=True):
            if s is None:
                return 0
            words = re.sub("[^A-Za-z0-9 ]+", "", s).split()
            for word_ in words:
                word = word_.lower()
                if lemmatisize:
                    word = self.lem.lemmatize(word)
                if word in words_list:
                    return 1
            return 0

        return pd.DataFrame(
            dict(
                [
                    (
                        self.column_name,
                        data[self.column_name].apply(
                            lambda x: listed_words_present(
                                x,
                                words_list=self.words_list,
                                lemmatisize=self.lemmatisize,
                            )
                        ),
                    )
                ]
            )
        )



import re
from typing import Optional

import pandas as pd
from nltk.corpus import words
from nltk.stem.wordnet import WordNetLemmatizer

from evidently.base_metric import additional_feature
from evidently.features.generated_features import GeneratedFeature
from evidently.utils.data_preprocessing import DataDefinition


class OOVWordsPercentage(GeneratedFeature):
    def __init__(self, column_name: str, ignore_words=()):
        self.lem = WordNetLemmatizer()
        self.eng_words = set(words.words())
        self.column_name = column_name
        self.ignore_words = ignore_words

    def generate_feature(self, data: pd.DataFrame, data_definition: DataDefinition) -> pd.DataFrame:
        def oov_share(s, ignore_words=()):
            if s is None:
                return 0
            oov_num = 0
            words_ = re.sub("[^A-Za-z0-9 ]+", "", s).split()  # leave only letters, digits and spaces, split by spaces
            for word in words_:
                if word.lower() not in ignore_words and self.lem.lemmatize(word.lower()) not in self.eng_words:
                    oov_num += 1
            return 100 * oov_num / len(words_)

        return pd.DataFrame(
            dict(
                [
                    (
                        self.column_name,
                        data[self.column_name].apply(lambda x: oov_share(x, ignore_words=self.ignore_words)),
                    )
                ]
            )
        )

    def feature_name(self):
        return additional_feature(self, self.column_name)

    def get_parameters(self) -> Optional[tuple]:
        return self.column_name, self.ignore_words



import abc
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Optional
from typing import Type

import pandas

from evidently.analyzers.base_analyzer import Analyzer
from evidently.pipeline.column_mapping import ColumnMapping
from evidently.pipeline.stage import PipelineStage


class ProfileSection(PipelineStage):
    _result: Optional[dict]

    @abc.abstractmethod
    def analyzers(self) -> Iterable[Type[Analyzer]]:
        raise NotImplementedError()

    @abc.abstractmethod
    def part_id(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def calculate(
        self,
        reference_data: pandas.DataFrame,
        current_data: Optional[pandas.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results: Dict[Type[Analyzer], Any],
    ) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_results(self) -> Optional[dict]:
        raise NotImplementedError()



import dataclasses
from abc import ABC
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
from pandas.core.dtypes.common import infer_dtype_from_object

from evidently.metrics import ColumnRegExpMetric
from evidently.metrics import ColumnSummaryMetric
from evidently.metrics import DatasetMissingValuesMetric
from evidently.metrics import DatasetSummaryMetric
from evidently.metrics.data_integrity.dataset_missing_values_metric import DatasetMissingValues
from evidently.metrics.data_integrity.dataset_missing_values_metric import DatasetMissingValuesMetricResult
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import DetailsInfo
from evidently.renderers.base_renderer import TestHtmlInfo
from evidently.renderers.base_renderer import TestRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.tests.base_test import BaseCheckValueTest
from evidently.tests.base_test import GroupData
from evidently.tests.base_test import GroupingTypes
from evidently.tests.base_test import Test
from evidently.tests.base_test import TestResult
from evidently.tests.base_test import TestValueCondition
from evidently.tests.utils import approx
from evidently.tests.utils import dataframes_to_table
from evidently.tests.utils import plot_dicts_to_table
from evidently.tests.utils import plot_value_counts_tables_ref_curr
from evidently.utils.data_operations import DatasetColumns
from evidently.utils.generators import BaseGenerator
from evidently.utils.types import Numeric

DATA_INTEGRITY_GROUP = GroupData("data_integrity", "Data Integrity", "")
GroupingTypes.TestGroup.add_value(DATA_INTEGRITY_GROUP)


class BaseIntegrityValueTest(BaseCheckValueTest, ABC):
    group = DATA_INTEGRITY_GROUP.id
    metric: DatasetSummaryMetric

    def __init__(
        self,
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        super().__init__(eq=eq, gt=gt, gte=gte, is_in=is_in, lt=lt, lte=lte, not_eq=not_eq, not_in=not_in)
        self.metric = DatasetSummaryMetric()


class TestNumberOfColumns(BaseIntegrityValueTest):
    """Number of all columns in the data, including utility columns (id/index, datetime, target, predictions)"""

    name = "Number of Columns"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_stats = self.metric.get_result().reference

        if reference_stats is not None:
            return TestValueCondition(eq=reference_stats.number_of_columns)

        return TestValueCondition(gt=0)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.number_of_columns

    def get_description(self, value: Numeric) -> str:
        return f"The number of columns is {value}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestNumberOfColumns)
class TestNumberOfColumnsRenderer(TestRenderer):
    def render_json(self, obj: TestNumberOfColumns) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["number_of_columns"] = obj.value
        return base

    def render_html(self, obj: TestNumberOfColumns) -> TestHtmlInfo:
        info = super().render_html(obj)
        columns = ["column name", "current dtype"]
        dict_curr = obj.metric.get_result().current.columns_type
        dict_ref = None
        reference_stats = obj.metric.get_result().reference

        if reference_stats is not None:
            dict_ref = reference_stats.columns_type
            columns = columns + ["reference dtype"]

        additional_plots = plot_dicts_to_table(dict_curr, dict_ref, columns, "number_of_column", "diff")
        info.details = additional_plots
        return info


class TestNumberOfRows(BaseIntegrityValueTest):
    """Number of rows in the data"""

    name = "Number of Rows"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_stats = self.metric.get_result().reference

        if reference_stats is not None:
            return TestValueCondition(eq=approx(reference_stats.number_of_rows, relative=0.1))

        return TestValueCondition(gt=30)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.number_of_rows

    def get_description(self, value: Numeric) -> str:
        return f"The number of rows is {value}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestNumberOfRows)
class TestNumberOfRowsRenderer(TestRenderer):
    def render_json(self, obj: TestNumberOfRows) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["number_of_rows"] = obj.value
        return base


class BaseIntegrityMissingValuesValuesTest(BaseCheckValueTest, ABC):
    group = DATA_INTEGRITY_GROUP.id
    metric: DatasetMissingValuesMetric

    def __init__(
        self,
        missing_values: Optional[list] = None,
        replace: bool = True,
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        super().__init__(eq=eq, gt=gt, gte=gte, is_in=is_in, lt=lt, lte=lte, not_eq=not_eq, not_in=not_in)
        self.metric = DatasetMissingValuesMetric(missing_values=missing_values, replace=replace)


class BaseTestMissingValuesRenderer(TestRenderer):
    """Common class for tests of missing values.
    Some tests have the same details visualizations.
    """

    MISSING_VALUES_NAMING_MAPPING = {
        None: "Pandas nulls (None, NAN, etc.)",
        "": '"" (empty string)',
        np.inf: 'Numpy "inf" value',
        -np.inf: 'Numpy "-inf" value',
    }

    @staticmethod
    def _get_number_and_percents_of_missing_values(missing_values_info: DatasetMissingValues) -> pd.DataFrame:
        """Get a string with missing values numbers and percents from info for results table"""
        result = {}

        for columns_name in missing_values_info.number_of_missing_values_by_column:
            missing_values_count = missing_values_info.number_of_missing_values_by_column[columns_name]
            percent_count = missing_values_info.share_of_missing_values_by_column[columns_name] * 100
            result[columns_name] = f"{missing_values_count} ({percent_count:.2f}%)"

        return pd.DataFrame.from_dict(
            {
                name: dict(
                    value=missing_values_info.number_of_missing_values_by_column[name],
                    display=f"{missing_values_info.number_of_missing_values_by_column[name]}"
                    f" ({missing_values_info.share_of_missing_values_by_column[name] * 100:.2f}%)",
                )
                for name in missing_values_info.number_of_missing_values_by_column.keys()
            },
            orient="index",
            columns=["value", "display"],
        )

    def get_table_with_missing_values_and_percents_by_column(
        self, info: TestHtmlInfo, metric_result: DatasetMissingValuesMetricResult, name: str
    ) -> TestHtmlInfo:
        """Get a table with missing values number and percents"""
        columns = ["column name", "current number of missing values"]
        dict_curr = self._get_number_and_percents_of_missing_values(metric_result.current)
        dict_ref = None
        reference_stats = metric_result.reference

        if reference_stats is not None:
            # add one more column and values for reference data
            columns.append("reference number of missing values")
            dict_ref = self._get_number_and_percents_of_missing_values(reference_stats)

        additional_plots = dataframes_to_table(dict_curr, dict_ref, columns, name)
        info.details = additional_plots
        return info

    def _replace_missing_values_to_description(self, values: dict) -> dict:
        """Replace missing values in the dict keys to human-readable string"""
        return {self.MISSING_VALUES_NAMING_MAPPING.get(k, k): v for k, v in values.items()}

    def get_table_with_number_of_missing_values_by_one_missing_value(
        self, info: TestHtmlInfo, current_missing_values: dict, reference_missing_values: Optional[dict], name: str
    ) -> TestHtmlInfo:
        columns = ["missing value", "current number of missing values"]
        dict_curr = self._replace_missing_values_to_description(current_missing_values)
        dict_ref: Optional[dict] = None

        if reference_missing_values is not None:
            # add one more column and values for reference data
            columns.append("reference number of missing values")
            # cast keys to str because None could be in keys, and it is not processed correctly in visual tables
            dict_ref = self._replace_missing_values_to_description(reference_missing_values)

        additional_plots = plot_dicts_to_table(dict_curr, dict_ref, columns, name)
        info.details = additional_plots
        return info


class TestNumberOfDifferentMissingValues(BaseIntegrityMissingValuesValuesTest):
    """Check a number of different encoded missing values."""

    name = "Different Types of Missing Values"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_missing_values = self.metric.get_result().reference

        if reference_missing_values is not None:
            return TestValueCondition(eq=reference_missing_values.number_of_different_missing_values)

        return TestValueCondition(eq=0)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.number_of_different_missing_values

    def get_description(self, value: Numeric) -> str:
        return (
            f"The number of differently encoded types of missing values is {value}. "
            f"The test threshold is {self.get_condition()}."
        )


@default_renderer(wrap_type=TestNumberOfDifferentMissingValues)
class TestNumberOfDifferentMissingValuesRenderer(BaseTestMissingValuesRenderer):
    def render_json(self, obj: TestNumberOfDifferentMissingValues) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["number_of_different_missing_values"] = obj.value
        return base

    def render_html(self, obj: TestNumberOfDifferentMissingValues) -> TestHtmlInfo:
        """Get a table with a missing value and number of the value in the dataset"""
        info = super().render_html(obj)
        metric_result = obj.metric.get_result()
        current_missing_values = metric_result.current.different_missing_values

        if metric_result.reference is None:
            reference_missing_values = None

        else:
            reference_missing_values = metric_result.reference.different_missing_values

        return self.get_table_with_number_of_missing_values_by_one_missing_value(
            info, current_missing_values, reference_missing_values, "number_of_different_missing_values"
        )


class TestNumberOfMissingValues(BaseIntegrityMissingValuesValuesTest):
    """Check a number of missing values."""

    name = "The Number of Missing Values"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_missing_values = self.metric.get_result().reference

        if reference_missing_values is not None:
            curr_number_of_rows = self.metric.get_result().current.number_of_rows
            ref_number_of_rows = reference_missing_values.number_of_rows
            mult = curr_number_of_rows / ref_number_of_rows
            return TestValueCondition(
                lte=approx(reference_missing_values.number_of_missing_values * mult, relative=0.1)
            )

        return TestValueCondition(eq=0)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.number_of_missing_values

    def get_description(self, value: Numeric) -> str:
        return f"The number of missing values is {value}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestNumberOfMissingValues)
class TestNumberOfMissingValuesRenderer(BaseTestMissingValuesRenderer):
    def render_json(self, obj: TestNumberOfMissingValues) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["number_of_missing_values"] = obj.value
        return base

    def render_html(self, obj: TestNumberOfMissingValues) -> TestHtmlInfo:
        info = super().render_html(obj)
        metric_result = obj.metric.get_result()
        return self.get_table_with_missing_values_and_percents_by_column(
            info, metric_result, "number_of_missing_values"
        )


class TestShareOfMissingValues(BaseIntegrityMissingValuesValuesTest):
    """Check a share of missing values."""

    name = "Share of Missing Values"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference = self.metric.get_result().reference

        if reference is not None:
            return TestValueCondition(lte=approx(reference.share_of_missing_values, relative=0.1))

        return TestValueCondition(eq=0)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.share_of_missing_values

    def get_description(self, value: Numeric) -> str:
        return f"The share of missing values is {value:.3g}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestShareOfMissingValues)
class TestShareOfMissingValuesRenderer(BaseTestMissingValuesRenderer):
    def render_json(self, obj: TestShareOfMissingValues) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["share_of_missing_values"] = obj.value
        return base

    def render_html(self, obj: TestNumberOfMissingValues) -> TestHtmlInfo:
        info = super().render_html(obj)
        metric_result = obj.metric.get_result()
        return self.get_table_with_missing_values_and_percents_by_column(info, metric_result, "share_of_missing_values")


class TestNumberOfColumnsWithMissingValues(BaseIntegrityMissingValuesValuesTest):
    """Check a number of columns with a missing value."""

    name = "The Number of Columns With Missing Values"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference = self.metric.get_result().reference

        if reference is not None:
            return TestValueCondition(lte=reference.number_of_columns_with_missing_values)

        return TestValueCondition(eq=0)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.number_of_columns_with_missing_values

    def get_description(self, value: Numeric) -> str:
        return (
            f"The number of columns with missing values is {value}. " f"The test threshold is {self.get_condition()}."
        )


@default_renderer(wrap_type=TestNumberOfColumnsWithMissingValues)
class TestNumberOfColumnsWithMissingValuesRenderer(BaseTestMissingValuesRenderer):
    def render_json(self, obj: TestNumberOfColumnsWithMissingValues) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["number_of_columns_with_missing_values"] = obj.value
        return base

    def render_html(self, obj: TestNumberOfMissingValues) -> TestHtmlInfo:
        info = super().render_html(obj)
        metric_result = obj.metric.get_result()
        return self.get_table_with_missing_values_and_percents_by_column(
            info, metric_result, "number_of_columns_with_missing_values"
        )


class TestShareOfColumnsWithMissingValues(BaseIntegrityMissingValuesValuesTest):
    """Check a share of columns with a missing value."""

    name = "The Share of Columns With Missing Values"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference = self.metric.get_result().reference

        if reference is not None:
            return TestValueCondition(lte=reference.share_of_columns_with_missing_values)

        return TestValueCondition(eq=0)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.share_of_columns_with_missing_values

    def get_description(self, value: Numeric) -> str:
        return (
            f"The share of columns with missing values is {value:.3g}. "
            f"The test threshold is {self.get_condition()}."
        )


@default_renderer(wrap_type=TestShareOfColumnsWithMissingValues)
class TestShareOfColumnsWithMissingValuesRenderer(BaseTestMissingValuesRenderer):
    def render_json(self, obj: TestShareOfColumnsWithMissingValues) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["share_of_columns_with_missing_values"] = obj.value
        return base

    def render_html(self, obj: TestNumberOfMissingValues) -> TestHtmlInfo:
        info = super().render_html(obj)
        metric_result = obj.metric.get_result()
        return self.get_table_with_missing_values_and_percents_by_column(
            info, metric_result, "share_of_columns_with_missing_values"
        )


class TestNumberOfRowsWithMissingValues(BaseIntegrityMissingValuesValuesTest):
    """Check a number of rows with a missing value."""

    name = "The Number Of Rows With Missing Values"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference = self.metric.get_result().reference

        if reference is not None:
            curr_number_of_rows = self.metric.get_result().current.number_of_rows
            ref_number_of_rows = reference.number_of_rows
            mult = curr_number_of_rows / ref_number_of_rows
            return TestValueCondition(lte=approx(reference.number_of_rows_with_missing_values * mult, relative=0.1))

        return TestValueCondition(eq=0)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.number_of_rows_with_missing_values

    def get_description(self, value: Numeric) -> str:
        return f"The number of rows with missing values is {value}. " f"The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestNumberOfRowsWithMissingValues)
class TestNumberOfRowsWithMissingValuesRenderer(BaseTestMissingValuesRenderer):
    def render_json(self, obj: TestNumberOfRowsWithMissingValues) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["number_of_rows_with_missing_values"] = obj.value
        return base


class TestShareOfRowsWithMissingValues(BaseIntegrityMissingValuesValuesTest):
    """Check a share of rows with a missing value."""

    name = "The Share of Rows With Missing Values"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference = self.metric.get_result().reference

        if reference is not None:
            return TestValueCondition(lte=approx(reference.share_of_rows_with_missing_values, relative=0.1))

        return TestValueCondition(eq=0)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.share_of_rows_with_missing_values

    def get_description(self, value: Numeric) -> str:
        return (
            f"The share of rows with missing values is {value:.3g}. " f"The test threshold is {self.get_condition()}."
        )


@default_renderer(wrap_type=TestShareOfRowsWithMissingValues)
class TestShareOfRowsWithMissingValuesRenderer(BaseTestMissingValuesRenderer):
    def render_json(self, obj: TestShareOfRowsWithMissingValues) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["share_of_rows_with_missing_values"] = obj.value
        return base


class BaseIntegrityColumnMissingValuesTest(BaseCheckValueTest, ABC):
    group = DATA_INTEGRITY_GROUP.id
    metric: DatasetMissingValuesMetric
    column_name: str

    def __init__(
        self,
        column_name: str,
        missing_values: Optional[list] = None,
        replace: bool = True,
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        super().__init__(eq=eq, gt=gt, gte=gte, is_in=is_in, lt=lt, lte=lte, not_eq=not_eq, not_in=not_in)
        self.column_name = column_name
        self.metric = DatasetMissingValuesMetric(missing_values=missing_values, replace=replace)


class TestColumnNumberOfDifferentMissingValues(BaseIntegrityColumnMissingValuesTest):
    """Check a number of differently encoded missing values in one column."""

    name = "Different Types of Missing Values in a Column"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_missing_values = self.metric.get_result().reference

        if reference_missing_values is not None:
            if self.column_name not in reference_missing_values.number_of_different_missing_values_by_column:
                raise ValueError(
                    f"Cannot define test default conditions: no column '{self.column_name}' in reference dataset."
                )

            ref_value = reference_missing_values.number_of_different_missing_values_by_column[self.column_name]
            return TestValueCondition(lte=ref_value)

        return TestValueCondition(eq=0)

    def calculate_value_for_test(self) -> Numeric:
        metric_data = self.metric.get_result().current
        return metric_data.number_of_different_missing_values_by_column[self.column_name]

    def get_description(self, value: Numeric) -> str:
        return (
            f"The number of differently encoded types of missing values in the column **{self.column_name}** "
            f"is {value}. The test threshold is {self.get_condition()}."
        )


@default_renderer(wrap_type=TestColumnNumberOfDifferentMissingValues)
class TestColumnNumberOfDifferentMissingValuesRenderer(BaseTestMissingValuesRenderer):
    def render_json(self, obj: TestColumnNumberOfDifferentMissingValues) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["number_of_different_missing_values"] = obj.value
        base["parameters"]["column_name"] = obj.column_name
        return base

    def render_html(self, obj: TestColumnNumberOfDifferentMissingValues) -> TestHtmlInfo:
        """Get a table with a missing value and number of the value in the dataset"""
        info = super().render_html(obj)
        metric_result = obj.metric.get_result()
        current_missing_values = metric_result.current.different_missing_values_by_column[obj.column_name]

        if metric_result.reference is None:
            reference_missing_values = None

        else:
            reference_missing_values = metric_result.reference.different_missing_values_by_column[obj.column_name]

        return self.get_table_with_number_of_missing_values_by_one_missing_value(
            info, current_missing_values, reference_missing_values, "number_of_different_missing_values"
        )


class TestColumnNumberOfMissingValues(BaseIntegrityColumnMissingValuesTest):
    """Check a number of missing values in one column."""

    name = "The Number of Missing Values in a Column"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_missing_values = self.metric.get_result().reference

        if reference_missing_values is not None:
            curr_number_of_rows = self.metric.get_result().current.number_of_rows
            ref_number_of_rows = reference_missing_values.number_of_rows
            mult = curr_number_of_rows / ref_number_of_rows
            ref_value = reference_missing_values.number_of_missing_values_by_column[self.column_name]
            return TestValueCondition(lte=approx(ref_value * mult, relative=0.1))

        return TestValueCondition(eq=0)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.number_of_missing_values_by_column[self.column_name]

    def get_description(self, value: Numeric) -> str:
        return (
            f"The number of missing values in the column **{self.column_name}** is {value}. "
            f"The test threshold is {self.get_condition()}."
        )


@default_renderer(wrap_type=TestColumnNumberOfMissingValues)
class TestColumnNumberOfMissingValuesRenderer(BaseTestMissingValuesRenderer):
    def render_json(self, obj: TestColumnNumberOfMissingValues) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["number_of_missing_values"] = obj.value
        base["parameters"]["column_name"] = obj.column_name
        return base


class TestColumnShareOfMissingValues(BaseIntegrityColumnMissingValuesTest):
    """Check a share of missing values in one column."""

    name = "The Share of Missing Values in a Column"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference = self.metric.get_result().reference

        if reference is not None:
            ref_value = reference.share_of_missing_values_by_column[self.column_name]
            return TestValueCondition(lte=approx(ref_value, relative=0.1))

        return TestValueCondition(eq=0)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.share_of_missing_values_by_column[self.column_name]

    def get_description(self, value: Numeric) -> str:
        return (
            f"The share of missing values in the column **{self.column_name}** is {value:.3g}. "
            f"The test threshold is {self.get_condition()}."
        )


class TestAllColumnsShareOfMissingValues(BaseGenerator):
    columns: Optional[List[str]]

    def __init__(self, columns: Optional[List[str]] = None):
        self.columns = columns

    def generate(self, columns_info: DatasetColumns) -> List[TestColumnShareOfMissingValues]:
        if self.columns is None:
            columns = columns_info.get_all_columns_list()

        else:
            columns = self.columns

        return [TestColumnShareOfMissingValues(column_name=name) for name in columns]


@default_renderer(wrap_type=TestColumnShareOfMissingValues)
class TestColumnShareOfMissingValuesRenderer(BaseTestMissingValuesRenderer):
    def render_json(self, obj: TestColumnShareOfMissingValues) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["share_of_missing_values"] = obj.value
        base["parameters"]["column_name"] = obj.column_name
        return base


class TestNumberOfConstantColumns(BaseIntegrityValueTest):
    """Number of columns contained only one unique value"""

    name = "Number of Constant Columns"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_stats = self.metric.get_result().reference

        if reference_stats is not None:
            value = reference_stats.number_of_constant_columns
            return TestValueCondition(lte=value)

        return TestValueCondition(eq=0)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.number_of_constant_columns

    def get_description(self, value: Numeric) -> str:
        return f"The number of constant columns is {value}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestNumberOfConstantColumns)
class TestNumberOfConstantColumnsRenderer(TestRenderer):
    def render_json(self, obj: TestNumberOfConstantColumns) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["number_of_constant_columns"] = obj.value
        return base

    def render_html(self, obj: TestNumberOfConstantColumns) -> TestHtmlInfo:
        info = super().render_html(obj)
        columns = ["column name", "current nunique"]
        dict_curr = obj.metric.get_result().current.number_uniques_by_columns
        dict_ref = {}
        reference_stats = obj.metric.get_result().reference

        if reference_stats is not None:
            dict_ref = reference_stats.number_uniques_by_columns
            columns = columns + ["reference nunique"]

        additional_plots = plot_dicts_to_table(dict_curr, dict_ref, columns, "number_of_constant_cols", "curr", True)
        info.details = additional_plots
        return info


class TestNumberOfEmptyRows(BaseIntegrityValueTest):
    """Number of rows contained all NAN values"""

    name = "Number of Empty Rows"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_stats = self.metric.get_result().reference

        if reference_stats is not None:
            ref_number_of_empty_rows = reference_stats.number_of_empty_rows
            curr_number_of_rows = self.metric.get_result().current.number_of_rows
            ref_number_of_rows = reference_stats.number_of_rows
            mult = curr_number_of_rows / ref_number_of_rows
            return TestValueCondition(eq=approx(ref_number_of_empty_rows * mult, 0.1))

        return TestValueCondition(eq=0)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.number_of_empty_rows

    def get_description(self, value: Numeric) -> str:
        return f"Number of Empty Rows is {value}. The test threshold is {self.get_condition()}."


class TestNumberOfEmptyColumns(BaseIntegrityValueTest):
    """Number of columns contained all NAN values"""

    name = "Number of Empty Columns"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_stats = self.metric.get_result().reference

        if reference_stats is not None:
            return TestValueCondition(lte=reference_stats.number_of_empty_columns)

        return TestValueCondition(eq=0)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.number_of_empty_columns

    def get_description(self, value: Numeric) -> str:
        return f"Number of Empty Columns is {value}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestNumberOfEmptyColumns)
class TestNumberOfEmptyColumnsRenderer(TestRenderer):
    def render_html(self, obj: TestNumberOfEmptyColumns) -> TestHtmlInfo:
        info = super().render_html(obj)
        columns = ["column name", "current number of NaNs"]
        dict_curr = obj.metric.get_result().current.nans_by_columns
        dict_ref = {}
        reference_stats = obj.metric.get_result().reference

        if reference_stats is not None:
            dict_ref = reference_stats.nans_by_columns
            columns = columns + ["reference number of NaNs"]

        additional_plots = plot_dicts_to_table(dict_curr, dict_ref, columns, "number_of_empty_columns")
        info.details = additional_plots
        return info


class TestNumberOfDuplicatedRows(BaseIntegrityValueTest):
    """How many rows have duplicates in the dataset"""

    name = "Number of Duplicate Rows"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_stats = self.metric.get_result().reference

        if reference_stats is not None:
            ref_num_of_duplicates = reference_stats.number_of_duplicated_rows
            curr_number_of_rows = self.metric.get_result().current.number_of_rows
            ref_number_of_rows = reference_stats.number_of_rows
            mult = curr_number_of_rows / ref_number_of_rows
            return TestValueCondition(eq=approx(ref_num_of_duplicates * mult, 0.1))

        return TestValueCondition(eq=0)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.number_of_duplicated_rows

    def get_description(self, value: Numeric) -> str:
        return f"The number of duplicate rows is {value}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestNumberOfDuplicatedRows)
class TestNumberOfDuplicatedRowsRenderer(TestRenderer):
    def render_json(self, obj: TestNumberOfDuplicatedRows) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["number_of_duplicated_rows"] = obj.value
        return base


class TestNumberOfDuplicatedColumns(BaseIntegrityValueTest):
    """How many columns have duplicates in the dataset"""

    name = "Number of Duplicate Columns"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        reference_stats = self.metric.get_result().reference

        if reference_stats is not None:
            value = reference_stats.number_of_duplicated_columns
            return TestValueCondition(lte=value)

        return TestValueCondition(eq=0)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().current.number_of_duplicated_columns

    def get_description(self, value: Numeric) -> str:
        return f"The number of duplicate columns is {value}. The test threshold is {self.get_condition()}."


@default_renderer(wrap_type=TestNumberOfDuplicatedColumns)
class TestNumberOfDuplicatedColumnsRenderer(TestRenderer):
    def render_json(self, obj: TestNumberOfDuplicatedColumns) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["number_of_duplicated_columns"] = obj.value
        return base


class BaseIntegrityByColumnsConditionTest(BaseCheckValueTest, ABC):
    group = DATA_INTEGRITY_GROUP.id
    data_integrity_metric: ColumnSummaryMetric
    column_name: str

    def __init__(
        self,
        column_name: str,
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        super().__init__(eq=eq, gt=gt, gte=gte, is_in=is_in, lt=lt, lte=lte, not_eq=not_eq, not_in=not_in)
        self.column_name = column_name
        self.data_integrity_metric = ColumnSummaryMetric(column_name=column_name)

    def groups(self) -> Dict[str, str]:
        if self.column_name is not None:
            return {GroupingTypes.ByFeature.id: self.column_name}
        return {}


class BaseIntegrityOneColumnTest(Test, ABC):
    group = DATA_INTEGRITY_GROUP.id
    metric: ColumnSummaryMetric
    column_name: str

    def __init__(self, column_name: str):
        self.column_name = column_name
        self.metric = ColumnSummaryMetric(column_name)

    def groups(self) -> Dict[str, str]:
        return {GroupingTypes.ByFeature.id: self.column_name}


class TestColumnAllConstantValues(BaseIntegrityOneColumnTest):
    """Test that there is only one unique value in a column"""

    name = "All Constant Values in a Column"
    metric: ColumnSummaryMetric

    def check(self):
        uniques_in_column = self.metric.get_result().current_characteristics.unique
        number_of_rows = self.metric.get_result().current_characteristics.number_of_rows
        column_name = self.column_name

        description = (
            f"The number of the unique values in the column **{column_name}** "
            f"is {uniques_in_column} out of {number_of_rows}"
        )

        if uniques_in_column <= 1:
            status = TestResult.FAIL

        else:
            status = TestResult.SUCCESS

        return TestResult(name=self.name, description=description, status=status, groups=self.groups())


@default_renderer(wrap_type=TestColumnAllConstantValues)
class TestColumnAllConstantValuesRenderer(TestRenderer):
    def render_html(self, obj: TestColumnAllConstantValues) -> TestHtmlInfo:
        info = super().render_html(obj)
        column_name = obj.column_name
        counts_data = obj.metric.get_result().plot_data.counts_of_values
        if counts_data is not None:
            curr_df = counts_data["current"]
            ref_df = None
            if "reference" in counts_data.keys():
                ref_df = counts_data["reference"]
            additional_plots = plot_value_counts_tables_ref_curr(column_name, curr_df, ref_df, "AllConstantValues")
            info.details = additional_plots
        return info


class TestColumnAllUniqueValues(BaseIntegrityOneColumnTest):
    """Test that there is only uniques values in a column"""

    name = "All Unique Values in a Column"

    def check(self):
        uniques_in_column = self.metric.get_result().current_characteristics.unique
        number_of_rows = self.metric.get_result().current_characteristics.number_of_rows
        nans_in_column = self.metric.get_result().current_characteristics.missing
        column_name = self.column_name

        description = (
            f"The number of the unique values in the column **{column_name}** "
            f"is {uniques_in_column}  out of {number_of_rows}"
        )

        if uniques_in_column != number_of_rows - nans_in_column:
            status = TestResult.FAIL

        else:
            status = TestResult.SUCCESS

        return TestResult(name=self.name, description=description, status=status, groups=self.groups())


@default_renderer(wrap_type=TestColumnAllUniqueValues)
class TestColumnAllUniqueValuesRenderer(TestRenderer):
    def render_html(self, obj: TestColumnAllUniqueValues) -> TestHtmlInfo:
        info = super().render_html(obj)
        column_name = obj.column_name
        counts_data = obj.metric.get_result().plot_data.counts_of_values
        if counts_data is not None:
            curr_df = counts_data["current"]
            ref_df = None
            if "reference" in counts_data.keys():
                ref_df = counts_data["reference"]
            additional_plots = plot_value_counts_tables_ref_curr(column_name, curr_df, ref_df, "AllUniqueValues")
            info.details = additional_plots
        return info


class TestColumnsType(Test):
    """This test compares columns type against the specified ones or a reference dataframe"""

    group = DATA_INTEGRITY_GROUP.id
    name = "Column Types"
    columns_type: Optional[dict]
    metric: DatasetSummaryMetric

    @dataclasses.dataclass
    class Result(TestResult):
        columns_types: Dict[str, Tuple[str, str]] = dataclasses.field(default_factory=dict)

    def __init__(self, columns_type: Optional[dict] = None):
        self.columns_type = columns_type
        self.metric = DatasetSummaryMetric()

    def check(self):
        status = TestResult.SUCCESS
        data_columns_type = self.metric.get_result().current.columns_type

        if self.columns_type is None:
            if self.metric.get_result().reference is None:
                status = TestResult.ERROR
                description = "Cannot compare column types without conditions or a reference"
                return TestResult(name=self.name, description=description, status=status)

            # get types from reference
            columns_type = self.metric.get_result().reference.columns_type

        else:
            columns_type = self.columns_type

            if not columns_type:
                status = TestResult.ERROR
                description = "Columns type condition is empty"
                return TestResult(name=self.name, description=description, status=status)

        invalid_types_count = 0
        columns_types = {}

        for column_name, expected_type_object in columns_type.items():
            real_column_type_object = data_columns_type.get(column_name)

            if real_column_type_object is None:
                status = TestResult.ERROR
                description = f"No column '{column_name}' in the metrics data"
                return TestResult(name=self.name, description=description, status=status)

            expected_type = infer_dtype_from_object(expected_type_object)
            real_column_type = infer_dtype_from_object(real_column_type_object)
            columns_types[column_name] = (real_column_type.__name__, expected_type.__name__)

            if expected_type == real_column_type or issubclass(real_column_type, expected_type):
                # types are matched or expected type is a parent
                continue

            status = TestResult.FAIL
            invalid_types_count += 1

        return self.Result(
            name=self.name,
            description=f"The number of columns with a type "
            f"mismatch is {invalid_types_count} out of {len(columns_type)}.",
            status=status,
            columns_types=columns_types,
        )


@default_renderer(wrap_type=TestColumnsType)
class TestColumnsTypeRenderer(TestRenderer):
    def render_json(self, obj: TestColumnsType) -> dict:
        base = super().render_json(obj)
        base["parameters"]["columns"] = [
            dict(
                column_name=column_name,
                actual_type=types[0],
                expected_type=types[1],
            )
            for column_name, types in obj.get_result().columns_types.items()
        ]
        return base

    def render_html(self, obj: TestColumnsType) -> TestHtmlInfo:
        info = super().render_html(obj)

        info.details = [
            DetailsInfo(
                title="",
                info=BaseWidgetInfo(
                    title="",
                    type="table",
                    params={
                        "header": ["Column Name", "Actual Type", "Expected Type"],
                        "data": [
                            [column_name, *types] for column_name, types in obj.get_result().columns_types.items()
                        ],
                    },
                    size=2,
                ),
            ),
        ]
        return info


class TestColumnRegExp(BaseCheckValueTest, ABC):
    group = DATA_INTEGRITY_GROUP.id
    name = "RegExp Match"
    metric: ColumnRegExpMetric
    column_name: str

    def __init__(
        self,
        column_name: str,
        reg_exp: str,
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
    ):
        super().__init__(eq=eq, gt=gt, gte=gte, is_in=is_in, lt=lt, lte=lte, not_eq=not_eq, not_in=not_in)
        self.column_name = column_name
        self.metric = ColumnRegExpMetric(column_name=column_name, reg_exp=reg_exp)

    def groups(self) -> Dict[str, str]:
        if self.column_name is not None:
            return {GroupingTypes.ByFeature.id: self.column_name}
        return {}

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition

        metric_result = self.metric.get_result()

        if metric_result.reference:
            ref_value = metric_result.reference.number_of_not_matched
            mult = metric_result.current.number_of_rows / metric_result.reference.number_of_rows

            if mult is not None:
                return TestValueCondition(eq=approx(ref_value * mult, relative=0.1))

        return TestValueCondition(eq=0)

    def calculate_value_for_test(self) -> Optional[Numeric]:
        return self.metric.get_result().current.number_of_not_matched

    def get_description(self, value: Numeric) -> str:
        return (
            f"The number of the mismatched values in the column **{self.column_name}** is {value}. "
            f"The test threshold is {self.get_condition()}."
        )


@default_renderer(wrap_type=TestColumnRegExp)
class TestColumnRegExpRenderer(TestRenderer):
    def render_html(self, obj: TestColumnRegExp) -> TestHtmlInfo:
        info = super().render_html(obj)
        column_name = obj.column_name
        metric_result = obj.metric.get_result()

        if metric_result.current.table_of_not_matched:
            curr_df = pd.DataFrame(metric_result.current.table_of_not_matched.items())
            curr_df.columns = ["x", "count"]

        else:
            curr_df = pd.DataFrame(columns=["x", "count"])

        ref_df = None

        if metric_result.reference is not None and metric_result.reference.table_of_not_matched:
            ref_df = pd.DataFrame(metric_result.reference.table_of_not_matched.items())
            ref_df.columns = ["x", "count"]

        additional_plots = plot_value_counts_tables_ref_curr(
            column_name, curr_df, ref_df, f"{column_name}_ColumnValueRegExp"
        )
        info.details = additional_plots
        return info



#!/usr/bin/env python
# coding: utf-8
from typing import Optional

import pandas as pd

from evidently import ColumnMapping
from evidently.analyzers.classification_performance_analyzer import (
    ClassificationPerformanceAnalyzer,
)
from evidently.analyzers.prob_classification_performance_analyzer import (
    ProbClassificationPerformanceAnalyzer,
)
from evidently.analyzers.regression_performance_analyzer import (
    RegressionPerformanceAnalyzer,
)
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo


class TargetNameWidget(Widget):
    def __init__(self, title: str, kind: str):
        super().__init__(title)
        self.kind = kind  # regression, classification or prob_classification

    def analyzers(self):
        if self.kind == "regression":
            return [RegressionPerformanceAnalyzer]
        if self.kind == "classification":
            return [ClassificationPerformanceAnalyzer]
        if self.kind == "prob_classification":
            return [ProbClassificationPerformanceAnalyzer]
        raise ValueError(f"Unexpected kind({self.kind}) of TargetNameWidget")

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:

        results_columns = None

        if self.kind == "regression":
            regression_results = RegressionPerformanceAnalyzer.get_results(
                analyzers_results
            )

            if regression_results:
                results_columns = regression_results.columns

        elif self.kind == "classification":
            classification_results = ClassificationPerformanceAnalyzer.get_results(
                analyzers_results
            )

            if classification_results:
                results_columns = classification_results.columns

        elif self.kind == "prob_classification":
            prob_classification_results = (
                ProbClassificationPerformanceAnalyzer.get_results(analyzers_results)
            )

            if prob_classification_results:
                results_columns = prob_classification_results.columns

        else:
            raise ValueError(f"Unexpected kind({self.kind}) of TagetNameWidget")

        if not results_columns:
            raise ValueError(f"Widget [{self.title}]: analyzer results not found")

        if (
            results_columns.utility_columns.target is None
            or results_columns.utility_columns.prediction is None
        ):
            raise ValueError(
                f"Widget [{self.title}] requires 'target' and 'prediction' columns"
            )

        return BaseWidgetInfo(
            title="",
            type="counter",
            size=2,
            params={
                "counters": [
                    {
                        "value": "",
                        "label": self.title
                        + " Target: '"
                        + results_columns.utility_columns.target
                        + "'",
                    }
                ]
            },
        )



import dataclasses
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import numpy as np
import pandas as pd

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.data_quality import get_rows_count
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import CounterData
from evidently.renderers.html_widgets import HistogramData
from evidently.renderers.html_widgets import TabData
from evidently.renderers.html_widgets import counter
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import histogram
from evidently.renderers.html_widgets import table_data
from evidently.renderers.html_widgets import widget_tabs


@dataclass
class DatasetMissingValues:
    """Statistics about missed values in a dataset"""

    # set of different missing values in the dataset
    different_missing_values: Dict[Any, int]
    # number of different missing values in the dataset
    number_of_different_missing_values: int
    # set of different missing values for each column
    different_missing_values_by_column: Dict[str, Dict[Any, int]]
    # count of different missing values for each column
    number_of_different_missing_values_by_column: Dict[str, int]
    # count of missing values in all dataset
    number_of_missing_values: int
    # share of missing values in all dataset
    share_of_missing_values: float
    # count of missing values for each column
    number_of_missing_values_by_column: Dict[str, int]
    # share of missing values for each column
    share_of_missing_values_by_column: Dict[str, float]
    # count of rows in the dataset
    number_of_rows: int
    # count of rows with a missing value
    number_of_rows_with_missing_values: int
    # share of rows with a missing value
    share_of_rows_with_missing_values: float
    # count of columns in the dataset
    number_of_columns: int
    # list of columns with a missing value
    columns_with_missing_values: List[str]
    # count of columns with a missing value
    number_of_columns_with_missing_values: int
    # share of columns with a missing value
    share_of_columns_with_missing_values: float


@dataclass
class DatasetMissingValuesMetricResult:
    current: DatasetMissingValues
    reference: Optional[DatasetMissingValues] = None


class DatasetMissingValuesMetric(Metric[DatasetMissingValuesMetricResult]):
    """Count missing values in a dataset.

    Missing value is a null or NaN value.

    Calculate an amount of missing values kinds and count for such values.
    NA-types like numpy.NaN, pandas.NaT are counted as one type.

    You can set you own missing values list with `missing_values` parameter.
    Value `None` in the list means that Pandas null values will be included in the calculation.

    If `replace` parameter is False - add defaults to user's list.
    If `replace` parameter is True - use values from `missing_values` list only.
    """

    # default missing values list
    DEFAULT_MISSING_VALUES = ["", np.inf, -np.inf, None]
    missing_values: frozenset

    def __init__(self, missing_values: Optional[list] = None, replace: bool = True) -> None:
        if missing_values is None:
            # use default missing values list if we have no user-defined values
            missing_values = self.DEFAULT_MISSING_VALUES

        elif not replace:
            # add default values to the user-defined list
            missing_values = self.DEFAULT_MISSING_VALUES + missing_values

        # use frozenset because metrics parameters should be immutable/hashable for deduplication
        self.missing_values = frozenset(missing_values)

    def _calculate_missing_values_stats(self, dataset: pd.DataFrame) -> DatasetMissingValues:
        different_missing_values = {value: 0 for value in self.missing_values}
        columns_with_missing_values = set()
        number_of_missing_values = 0
        number_of_missing_values_by_column: Dict[str, int] = {}
        different_missing_values_by_column: Dict[str, Dict[Any, int]] = {}

        for column_name in dataset.columns:
            number_of_missing_values_by_column[column_name] = 0
            different_missing_values_by_column[column_name] = {}

            for value in self.missing_values:
                different_missing_values_by_column[column_name][value] = 0

        number_of_rows_with_missing_values = 0
        number_of_columns = len(dataset.columns)
        number_of_rows = get_rows_count(dataset)

        for column_name in dataset.columns:
            # iterate by each value in custom missing values list and check the value in a column
            for missing_value in self.missing_values:
                if missing_value is None:
                    # check all pandas null-types like numpy.NAN, pandas.NA, pandas.NaT, etc
                    column_missing_value = dataset[column_name].isnull().sum()

                else:
                    column_missing_value = (dataset[column_name] == missing_value).sum()

                if column_missing_value > 0:
                    # increase overall counter
                    number_of_missing_values += column_missing_value
                    # increase by-column counter
                    number_of_missing_values_by_column[column_name] += column_missing_value
                    # increase by-missing-value counter for each column
                    different_missing_values_by_column[column_name][missing_value] += column_missing_value
                    # increase by-missing-value counter
                    different_missing_values[missing_value] += column_missing_value
                    # add the column to set of columns with a missing value
                    columns_with_missing_values.add(column_name)

        for _, row in dataset.iterrows():
            for missing_value in self.missing_values:
                if None in self.missing_values and row.isnull().any():
                    # check pandas null values
                    number_of_rows_with_missing_values += 1
                    break

                elif missing_value in row.values:
                    number_of_rows_with_missing_values += 1
                    break

        if number_of_rows == 0:
            share_of_missing_values_by_column = {}
            share_of_rows_with_missing_values = 0.0
            share_of_missing_values = 0.0

        else:
            share_of_missing_values_by_column = {
                column_name: value / number_of_rows for column_name, value in number_of_missing_values_by_column.items()
            }
            share_of_missing_values = number_of_missing_values / (number_of_columns * number_of_rows)
            share_of_rows_with_missing_values = number_of_rows_with_missing_values / number_of_rows

        number_of_different_missing_values_by_column = {}

        for column_name, missing_values in different_missing_values_by_column.items():
            # count a number of missing values that have a value in the column
            number_of_different_missing_values_by_column[column_name] = len(
                {keys for keys, values in missing_values.items() if values > 0}
            )

        number_of_columns_with_missing_values = len(columns_with_missing_values)
        number_of_different_missing_values = len(
            {k for k in different_missing_values if different_missing_values[k] > 0}
        )

        if number_of_columns == 0:
            share_of_columns_with_missing_values = 0.0

        else:
            share_of_columns_with_missing_values = number_of_columns_with_missing_values / number_of_columns

        return DatasetMissingValues(
            different_missing_values=different_missing_values,
            number_of_different_missing_values=number_of_different_missing_values,
            different_missing_values_by_column=different_missing_values_by_column,
            number_of_different_missing_values_by_column=number_of_different_missing_values_by_column,
            number_of_missing_values=number_of_missing_values,
            share_of_missing_values=share_of_missing_values,
            number_of_missing_values_by_column=number_of_missing_values_by_column,
            share_of_missing_values_by_column=share_of_missing_values_by_column,
            number_of_rows=number_of_rows,
            number_of_rows_with_missing_values=number_of_rows_with_missing_values,
            share_of_rows_with_missing_values=share_of_rows_with_missing_values,
            number_of_columns=number_of_columns,
            columns_with_missing_values=sorted(columns_with_missing_values),
            number_of_columns_with_missing_values=len(columns_with_missing_values),
            share_of_columns_with_missing_values=share_of_columns_with_missing_values,
        )

    def calculate(self, data: InputData) -> DatasetMissingValuesMetricResult:
        if not self.missing_values:
            raise ValueError("Missing values list should not be empty.")

        current_missing_values = self._calculate_missing_values_stats(data.current_data)

        if data.reference_data is not None:
            reference_missing_values: Optional[DatasetMissingValues] = self._calculate_missing_values_stats(
                data.reference_data
            )

        else:
            reference_missing_values = None

        return DatasetMissingValuesMetricResult(
            current=current_missing_values,
            reference=reference_missing_values,
        )


@default_renderer(wrap_type=DatasetMissingValuesMetric)
class DatasetMissingValuesMetricRenderer(MetricRenderer):
    def render_json(self, obj: DatasetMissingValuesMetric) -> dict:
        return dataclasses.asdict(obj.get_result().current)

    def _get_table_stat(self, dataset_name: str, stats: DatasetMissingValues) -> BaseWidgetInfo:
        matched_stat = [(k, v) for k, v in stats.number_of_missing_values_by_column.items()]
        matched_stat = sorted(matched_stat, key=lambda x: x[1], reverse=True)
        matched_stat_headers = ["Value", "Count"]
        table_tab = table_data(
            title="",
            column_names=matched_stat_headers,
            data=matched_stat,
        )
        histogram_tab = histogram(
            title="",
            primary_hist=HistogramData(
                name="",
                x=list(stats.number_of_missing_values_by_column.keys()),
                y=list(stats.number_of_missing_values_by_column.values()),
            ),
            color_options=self.color_options,
        )
        return widget_tabs(
            title=f"{dataset_name.capitalize()} dataset",
            tabs=[
                TabData(title="Table", widget=table_tab),
                TabData(
                    title="Histogram",
                    widget=histogram_tab,
                ),
            ],
        )

    @staticmethod
    def _get_info_string(stats: DatasetMissingValues) -> str:
        percents = round(stats.share_of_missing_values * 100, 3)
        return f"{stats.number_of_missing_values} ({percents}%)"

    def _get_overall_missing_values_info(self, metric_result: DatasetMissingValuesMetricResult) -> BaseWidgetInfo:
        counters = [
            CounterData.string("Missing values (Current data)", self._get_info_string(metric_result.current)),
        ]
        if metric_result.reference is not None:
            counters.append(
                CounterData.string("Missing values (Reference data)", self._get_info_string(metric_result.reference)),
            )

        return counter(
            title="",
            counters=counters,
        )

    def render_html(self, obj: DatasetMissingValuesMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        result = [
            header_text(label="Dataset Missing Values"),
            self._get_overall_missing_values_info(metric_result),
            self._get_table_stat(dataset_name="current", stats=metric_result.current),
        ]

        if metric_result.reference is not None:
            result.append(self._get_table_stat(dataset_name="reference", stats=metric_result.reference))

        return result



import dataclasses
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.features.non_letter_character_percentage_feature import NonLetterCharacterPercentage
from evidently.features.OOV_words_percentage_feature import OOVWordsPercentage
from evidently.features.text_length_feature import TextLength
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import WidgetSize
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import plotly_figure
from evidently.renderers.render_utils import get_distribution_plot_figure
from evidently.utils.data_operations import process_columns
from evidently.utils.data_operations import recognize_column_type
from evidently.utils.data_preprocessing import ColumnType
from evidently.utils.data_preprocessing import DataDefinition
from evidently.utils.visualizations import Distribution
from evidently.utils.visualizations import get_distribution_for_column


@dataclasses.dataclass
class TextDescriptorsDistributionResult:
    column_name: str
    current: Dict[str, Distribution]
    reference: Optional[Dict[str, Distribution]] = None


class TextDescriptorsDistribution(Metric[TextDescriptorsDistributionResult]):
    """Calculates distribution for the column"""

    column_name: str
    generated_text_features: Dict[str, Union[TextLength, NonLetterCharacterPercentage, OOVWordsPercentage]]

    def __init__(self, column_name: str) -> None:
        self.column_name = column_name
        self.generated_text_features = {}

    def required_features(self, data_definition: DataDefinition):
        column_type = data_definition.get_column(self.column_name).column_type
        if column_type == ColumnType.Text:
            self.generated_text_features = {}
            self.generated_text_features["Text Length"] = TextLength(self.column_name)
            self.generated_text_features["Non Letter Character %"] = NonLetterCharacterPercentage(self.column_name)
            self.generated_text_features["OOV %"] = OOVWordsPercentage(self.column_name)
            return list(self.generated_text_features.values())
        return []

    def get_parameters(self) -> tuple:
        return (self.column_name,)

    def calculate(self, data: InputData) -> TextDescriptorsDistributionResult:
        current_results = {}
        reference_results: Optional[Dict[str, Distribution]] = None
        if data.reference_data is not None:
            reference_results = {}
        if self.column_name not in data.current_data:
            raise ValueError(f"Column '{self.column_name}' was not found in current data.")

        if data.reference_data is not None:
            if self.column_name not in data.reference_data:
                raise ValueError(f"Column '{self.column_name}' was not found in reference data.")

        columns = process_columns(data.current_data, data.column_mapping)
        column_type = recognize_column_type(dataset=data.current_data, column_name=self.column_name, columns=columns)
        if column_type != "text":
            raise ValueError("Text column expected")
        for key, val in self.generated_text_features.items():
            current_column = data.get_current_column(val.feature_name())
            reference_column = None
            if data.reference_data is not None:
                reference_column = data.get_reference_column(val.feature_name())
            current, reference = get_distribution_for_column(
                column_type="num",
                current=current_column,
                reference=reference_column,
            )
            current_results[key] = current
            if reference_results is not None and reference is not None:
                reference_results[key] = reference

        return TextDescriptorsDistributionResult(
            column_name=self.column_name,
            current=current_results,
            reference=reference_results,
        )


@default_renderer(wrap_type=TextDescriptorsDistribution)
class TextDescriptorsDistributionRenderer(MetricRenderer):
    def render_json(self, obj: TextDescriptorsDistribution) -> dict:
        result = dataclasses.asdict(obj.get_result())
        result.pop("current", None)
        result.pop("reference", None)
        return result

    def render_html(self, obj: TextDescriptorsDistribution) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        result = [header_text(label=f"Distribution for column '{metric_result.column_name}'.")]
        for col in list(metric_result.current.keys()):
            reference = None
            if metric_result.reference is not None:
                reference = metric_result.reference[col]
            distr_fig = get_distribution_plot_figure(
                current_distribution=metric_result.current[col],
                reference_distribution=reference,
                color_options=self.color_options,
            )
            result.append(plotly_figure(title=col, figure=distr_fig, size=WidgetSize.FULL))

        return result



from collections import Counter

import numpy as np
import pandas as pd


def get_unique_not_nan_values_list_from_series(current_data: pd.Series, reference_data: pd.Series) -> list:
    """Get unique values from current and reference series, drop NaNs"""
    return list(set(reference_data.dropna().unique()) | set(current_data.dropna().unique()))


def get_binned_data(
    reference_data: pd.Series, current_data: pd.Series, feature_type: str, n: int, feel_zeroes: bool = True
):
    """Split variable into n buckets based on reference quantiles
    Args:
        reference_data: reference data
        current_data: current data
        feature_type: feature type
        n: number of quantiles
    Returns:
        reference_percents: % of records in each bucket for reference
        current_percents: % of records in each bucket for current
    """
    n_vals = reference_data.nunique()

    if feature_type == "num" and n_vals > 20:
        bins = np.histogram_bin_edges(list(reference_data) + list(current_data), bins="sturges")
        reference_percents = np.histogram(reference_data, bins)[0] / len(reference_data)
        current_percents = np.histogram(current_data, bins)[0] / len(current_data)

    else:
        keys = get_unique_not_nan_values_list_from_series(current_data=current_data, reference_data=reference_data)
        ref_feature_dict = {**dict.fromkeys(keys, 0), **dict(reference_data.value_counts())}
        current_feature_dict = {**dict.fromkeys(keys, 0), **dict(current_data.value_counts())}
        reference_percents = np.array([ref_feature_dict[key] / len(reference_data) for key in keys])
        current_percents = np.array([current_feature_dict[key] / len(current_data) for key in keys])

    if feel_zeroes:
        np.place(
            reference_percents,
            reference_percents == 0,
            min(reference_percents[reference_percents != 0]) / 10**6
            if min(reference_percents[reference_percents != 0]) <= 0.0001
            else 0.0001,
        )
        np.place(
            current_percents,
            current_percents == 0,
            min(current_percents[current_percents != 0]) / 10**6
            if min(current_percents[current_percents != 0]) <= 0.0001
            else 0.0001,
        )

    return reference_percents, current_percents


def permutation_test(reference_data, current_data, observed, test_statistic_func, iterations=100):

    """Perform a two-sided permutation test
    Args:
        reference_data: reference data
        current_data: current data
        observed: observed value
        test_statistic_func: the test statistic function
        iterations: number of times to permute
    Returns:
        p_value: two-sided p_value
    """
    np.random.seed(0)
    hold_test_statistic = []
    for i in range(iterations):
        combined_data = reference_data.tolist() + current_data.tolist()
        new_reference = np.random.choice(combined_data, len(reference_data), replace=False).tolist()
        count_combined = Counter(combined_data)
        count_new_reference = Counter(new_reference)
        new_current = list((count_combined - count_new_reference).elements())
        hold_test_statistic.append(test_statistic_func(pd.Series(new_reference), pd.Series(new_current)))

    p_val = sum(observed <= abs(np.array(hold_test_statistic))) / len(hold_test_statistic)
    return p_val


def generate_fisher2x2_contingency_table(reference_data: pd.Series, current_data: pd.Series) -> np.ndarray:
    """Generate 2x2 contingency matrix for fisher exact test
    Args:
        reference_data: reference data
        current_data: current data
    Raises:
        ValueError: if reference_data and current_data are not of equal length
    Returns:
        contingency_matrix: contingency_matrix for binary data
    """
    if reference_data.shape[0] != current_data.shape[0]:
        raise ValueError(
            "reference_data and current_data are not of equal length, please ensure that they are of equal length"
        )
    unique_categories = set(reference_data.unique().tolist() + current_data.unique().tolist())
    if len(unique_categories) != 2:
        unique_categories.add("placeholder")

    unique_categories = list(unique_categories)  # type: ignore
    unique_categories = dict(zip(unique_categories, [0, 1]))  # type: ignore

    reference_data = reference_data.map(unique_categories).values
    current_data = current_data.map(unique_categories).values

    zero_ref = reference_data.size - np.count_nonzero(reference_data)
    one_ref = np.count_nonzero(reference_data)

    zero_cur = current_data.size - np.count_nonzero(current_data)
    one_cur = np.count_nonzero(current_data)

    contingency_table = np.array([[one_cur, zero_cur], [one_ref, zero_ref]])

    return contingency_table



from typing import List
from typing import Optional

from evidently.base_metric import InputData
from evidently.metric_preset.metric_preset import MetricPreset
from evidently.metrics import ColumnSummaryMetric
from evidently.metrics import DatasetCorrelationsMetric
from evidently.metrics import DatasetSummaryMetric
from evidently.metrics.base_metric import generate_column_metrics
from evidently.metrics.data_integrity.dataset_missing_values_metric import DatasetMissingValuesMetric
from evidently.utils.data_operations import DatasetColumns


class DataQualityPreset(MetricPreset):
    """Metric preset for Data Quality analysis.

    Contains metrics:
    - DatasetSummaryMetric
    - ColumnSummaryMetric for each column
    - DatasetMissingValuesMetric
    - DatasetCorrelationsMetric

    Args:
        columns: list of columns for analysis.
    """

    columns: Optional[List[str]]

    def __init__(self, columns: Optional[List[str]] = None):
        super().__init__()
        self.columns = columns

    def generate_metrics(self, data: InputData, columns: DatasetColumns):
        return [
            DatasetSummaryMetric(),
            generate_column_metrics(ColumnSummaryMetric, columns=self.columns, skip_id_column=True),
            DatasetMissingValuesMetric(),
            DatasetCorrelationsMetric(),
        ]
